#
# Copyright (C) 2009-2010  Nexedi SA
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.

import unittest
from mock import Mock
from neo.tests import NeoUnitTestBase
from neo.lib.pt import PartitionTable
from neo.storage.app import Application
from neo.storage.handlers.verification import VerificationHandler
from neo.lib.protocol import CellStates, ErrorCodes
from neo.lib.exception import PrimaryFailure, OperationFailure
from neo.lib.util import p64, u64

class StorageVerificationHandlerTests(NeoUnitTestBase):

    def setUp(self):
        NeoUnitTestBase.setUp(self)
        self.prepareDatabase(number=1)
        # create an application object
        config = self.getStorageConfiguration(master_number=1)
        self.app = Application(config)
        self.verification = VerificationHandler(self.app)
        # define some variable to simulate client and storage node
        self.master_port = 10010
        self.storage_port = 10020
        self.client_port = 11011
        self.num_partitions = 1009
        self.num_replicas = 2
        self.app.operational = False
        self.app.load_lock_dict = {}
        self.app.pt = PartitionTable(self.num_partitions, self.num_replicas)

    def tearDown(self):
        self.app.close()
        del self.app
        super(StorageVerificationHandlerTests, self).tearDown()

    # Common methods
    def getLastUUID(self):
        return self.uuid

    def getClientConnection(self):
        address = ("127.0.0.1", self.client_port)
        return self.getFakeConnection(uuid=self.getNewUUID(), address=address)

    def getMasterConnection(self):
        return self.getFakeConnection(address=("127.0.0.1", self.master_port))

    # Tests
    def test_03_connectionClosed(self):
        conn = self.getClientConnection()
        self.app.listening_conn = object() # mark as running
        self.assertRaises(PrimaryFailure, self.verification.connectionClosed, conn,)
        # nothing happens
        self.checkNoPacketSent(conn)

    def test_07_askLastIDs(self):
        conn = self.getClientConnection()
        last_ptid = self.getPTID(1)
        last_oid = self.getOID(2)
        self.app.pt = Mock({'getID': last_ptid})
        class DummyDM(object):
            def getLastOID(self):
                raise KeyError
            getLastTID = getLastOID
        self.app.dm = DummyDM()
        self.verification.askLastIDs(conn)
        oid, tid, ptid = self.checkAnswerLastIDs(conn, decode=True)
        self.assertEqual(oid, None)
        self.assertEqual(tid, None)
        self.assertEqual(ptid, last_ptid)

        # return value stored in db
        conn = self.getClientConnection()
        self.app.dm = Mock({
            'getLastOID': last_oid,
            'getLastTID': p64(4),
        })
        self.verification.askLastIDs(conn)
        oid, tid, ptid = self.checkAnswerLastIDs(conn, decode=True)
        self.assertEqual(oid, last_oid)
        self.assertEqual(u64(tid), 4)
        self.assertEqual(ptid, self.app.pt.getID())
        call_list = self.app.dm.mockGetNamedCalls('getLastOID')
        self.assertEqual(len(call_list), 1)
        call_list[0].checkArgs()
        call_list = self.app.dm.mockGetNamedCalls('getLastTID')
        self.assertEqual(len(call_list), 1)
        call_list[0].checkArgs()

    def test_08_askPartitionTable(self):
        node = self.app.nm.createStorage(
            address=("127.7.9.9", 1),
            uuid=self.getNewUUID()
        )
        self.app.pt.setCell(1, node, CellStates.UP_TO_DATE)
        self.assertTrue(self.app.pt.hasOffset(1))
        conn = self.getClientConnection()
        self.verification.askPartitionTable(conn)
        ptid, row_list = self.checkAnswerPartitionTable(conn, decode=True)
        self.assertEqual(len(row_list), 1009)

    def test_10_notifyPartitionChanges(self):
        # old partition change
        conn = self.getMasterConnection()
        self.verification.notifyPartitionChanges(conn, 1, ())
        self.verification.notifyPartitionChanges(conn, 0, ())
        self.assertEqual(self.app.pt.getID(), 1)

        # new node
        conn = self.getMasterConnection()
        new_uuid = self.getNewUUID()
        cell = (0, new_uuid, CellStates.UP_TO_DATE)
        self.app.nm.createStorage(uuid=new_uuid)
        self.app.pt = PartitionTable(1, 1)
        self.app.dm = Mock({ })
        ptid, self.ptid = self.getTwoIDs()
        # pt updated
        self.verification.notifyPartitionChanges(conn, ptid, (cell, ))
        # check db update
        calls = self.app.dm.mockGetNamedCalls('changePartitionTable')
        self.assertEqual(len(calls), 1)
        self.assertEqual(calls[0].getParam(0), ptid)
        self.assertEqual(calls[0].getParam(1), (cell, ))

    def test_11_startOperation(self):
        conn = self.getMasterConnection()
        self.assertFalse(self.app.operational)
        self.verification.startOperation(conn)
        self.assertTrue(self.app.operational)

    def test_12_stopOperation(self):
        conn = self.getMasterConnection()
        self.assertRaises(OperationFailure, self.verification.stopOperation, conn)

    def test_13_askUnfinishedTransactions(self):
        # client connection with no data
        self.app.dm = Mock({
            'getUnfinishedTIDList': [],
        })
        conn = self.getMasterConnection()
        self.verification.askUnfinishedTransactions(conn)
        (max_tid, tid_list) = self.checkAnswerUnfinishedTransactions(conn, decode=True)
        self.assertEqual(len(tid_list), 0)
        call_list = self.app.dm.mockGetNamedCalls('getUnfinishedTIDList')
        self.assertEqual(len(call_list), 1)
        call_list[0].checkArgs()

        # client connection with some data
        self.app.dm = Mock({
            'getUnfinishedTIDList': [p64(4)],
        })
        conn = self.getMasterConnection()
        self.verification.askUnfinishedTransactions(conn)
        (max_tid, tid_list) = self.checkAnswerUnfinishedTransactions(conn, decode=True)
        self.assertEqual(len(tid_list), 1)
        self.assertEqual(u64(tid_list[0]), 4)

    def test_14_askTransactionInformation(self):
        # ask from client conn with no data
        self.app.dm = Mock({
            'getTransaction': None,
        })
        conn = self.getMasterConnection()
        tid = p64(1)
        self.verification.askTransactionInformation(conn, tid)
        code, message = self.checkErrorPacket(conn, decode=True)
        self.assertEqual(code, ErrorCodes.TID_NOT_FOUND)
        call_list = self.app.dm.mockGetNamedCalls('getTransaction')
        self.assertEqual(len(call_list), 1)
        call_list[0].checkArgs(tid, all=True)

        # input some tmp data and ask from client, must find both transaction
        self.app.dm = Mock({
            'getTransaction': ([p64(2)], 'u2', 'd2', 'e2', False),
        })
        conn = self.getClientConnection()
        self.verification.askTransactionInformation(conn, p64(1))
        tid, user, desc, ext, packed, oid_list = self.checkAnswerTransactionInformation(conn, decode=True)
        self.assertEqual(u64(tid), 1)
        self.assertEqual(user, 'u2')
        self.assertEqual(desc, 'd2')
        self.assertEqual(ext, 'e2')
        self.assertFalse(packed)
        self.assertEqual(len(oid_list), 1)
        self.assertEqual(u64(oid_list[0]), 2)

    def test_15_askObjectPresent(self):
        # client connection with no data
        self.app.dm = Mock({
            'objectPresent': False,
        })
        conn = self.getMasterConnection()
        oid, tid = p64(1), p64(2)
        self.verification.askObjectPresent(conn, oid, tid)
        code, message = self.checkErrorPacket(conn, decode=True)
        self.assertEqual(code, ErrorCodes.OID_NOT_FOUND)
        call_list = self.app.dm.mockGetNamedCalls('objectPresent')
        self.assertEqual(len(call_list), 1)
        call_list[0].checkArgs(oid, tid)

        # client connection with some data
        self.app.dm = Mock({
            'objectPresent': True,
        })
        conn = self.getMasterConnection()
        self.verification.askObjectPresent(conn, oid, tid)
        oid, tid = self.checkAnswerObjectPresent(conn, decode=True)
        self.assertEqual(u64(tid), 2)
        self.assertEqual(u64(oid), 1)

    def test_16_deleteTransaction(self):
        # client connection with no data
        self.app.dm = Mock({
            'deleteTransaction': None,
        })
        conn = self.getMasterConnection()
        oid_list = [self.getOID(1), self.getOID(2)]
        tid = p64(1)
        self.verification.deleteTransaction(conn, tid, oid_list)
        call_list = self.app.dm.mockGetNamedCalls('deleteTransaction')
        self.assertEqual(len(call_list), 1)
        call_list[0].checkArgs(tid, oid_list)

    def test_17_commitTransaction(self):
        # commit a transaction
        conn = self.getMasterConnection()
        dm = Mock()
        self.app.dm = dm
        self.verification.commitTransaction(conn, p64(1))
        self.assertEqual(len(dm.mockGetNamedCalls("finishTransaction")), 1)
        call = dm.mockGetNamedCalls("finishTransaction")[0]
        tid = call.getParam(0)
        self.assertEqual(u64(tid), 1)

if __name__ == "__main__":
    unittest.main()




#
# Copyright (c) 2011 Nexedi SARL and Contributors. All Rights Reserved.
#                    Julien Muchembled <jm@nexedi.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.

import threading
from thread import get_ident
from persistent import Persistent
from neo.storage.transactions import TransactionManager, \
    DelayedError, ConflictError
from neo.lib.connection import MTClientConnection
from neo.lib.protocol import NodeStates, Packets, ZERO_TID
from neo.tests.threaded import NEOCluster, NEOThreadedTest, \
    Patch, ConnectionFilter
from neo.client.pool import CELL_CONNECTED, CELL_GOOD

class PCounter(Persistent):
    value = 0

class PCounterWithResolution(PCounter):
    def _p_resolveConflict(self, old, saved, new):
        new['value'] += saved['value'] - old.get('value', 0)
        return new

class Test(NEOThreadedTest):

    def testDelayedUnlockInformation(self):
        except_list = []
        def delayUnlockInformation(conn, packet):
            return isinstance(packet, Packets.NotifyUnlockInformation)
        def onStoreObject(orig, tm, ttid, serial, oid, *args):
            if oid == resume_oid and delayUnlockInformation in master_storage:
                master_storage.remove(delayUnlockInformation)
            try:
                return orig(tm, ttid, serial, oid, *args)
            except Exception, e:
                except_list.append(e.__class__)
                raise
        cluster = NEOCluster(storage_count=1)
        try:
            cluster.start()
            t, c = cluster.getTransaction()
            c.root()[0] = ob = PCounter()
            master_storage = cluster.master.filterConnection(cluster.storage)
            try:
                resume_oid = None
                master_storage.add(delayUnlockInformation,
                    Patch(TransactionManager, storeObject=onStoreObject))
                t.commit()
                resume_oid = ob._p_oid
                ob._p_changed = 1
                t.commit()
                self.assertFalse(delayUnlockInformation in master_storage)
            finally:
                master_storage()
        finally:
            cluster.stop()
        self.assertEqual(except_list, [DelayedError])

    def _testDeadlockAvoidance(self, scenario):
        except_list = []
        delay = threading.Event(), threading.Event()
        ident = get_ident()
        def onStoreObject(orig, tm, ttid, serial, oid, *args):
            if oid == counter_oid:
                scenario[1] -= 1
                if not scenario[1]:
                    delay[0].set()
            try:
                return orig(tm, ttid, serial, oid, *args)
            except Exception, e:
                except_list.append(e.__class__)
                raise
        def onAsk(orig, conn, packet, *args, **kw):
            c2 = get_ident() == ident
            switch = isinstance(packet, Packets.AskBeginTransaction)
            if switch:
                if c2:
                    delay[1].wait()
            elif isinstance(packet, (Packets.AskStoreObject,
                                     Packets.AskFinishTransaction)):
                delay[c2].wait()
                scenario[0] -= 1
                switch = not scenario[0]
            try:
                return orig(conn, packet, *args, **kw)
            finally:
                if switch:
                    delay[c2].clear()
                    delay[1-c2].set()

        cluster = NEOCluster(storage_count=2, replicas=1)
        try:
            cluster.start()
            t, c = cluster.getTransaction()
            c.root()[0] = ob = PCounterWithResolution()
            t.commit()
            counter_oid = ob._p_oid
            del ob, t, c

            t1, c1 = cluster.getTransaction()
            t2, c2 = cluster.getTransaction()
            o1 = c1.root()[0]
            o2 = c2.root()[0]
            o1.value += 1
            o2.value += 2

            p = (Patch(TransactionManager, storeObject=onStoreObject),
                 Patch(MTClientConnection, ask=onAsk))
            try:
                t = self.newThread(t1.commit)
                t2.commit()
                t.join()
            finally:
                del p
            t1.begin()
            t2.begin()
            self.assertEqual(o1.value, 3)
            self.assertEqual(o2.value, 3)
        finally:
            cluster.stop()
        return except_list

    def testDelayedStore(self):
        # 0: C1 -> S1, S2
        # 1: C2 -> S1, S2 (delayed)
        # 2: C1 commits
        # 3: C2 resolves conflict
        self.assertEqual(self._testDeadlockAvoidance([2, 4]),
            [DelayedError, DelayedError, ConflictError, ConflictError])

    def testDeadlockAvoidance(self):
        # This test fail because deadlock avoidance is not fully implemented.
        # 0: C1 -> S1
        # 1: C2 -> S1, S2 (delayed)
        # 2: C1 -> S2 (deadlock)
        # 3: C2 commits
        # 4: C1 resolves conflict
        self.assertEqual(self._testDeadlockAvoidance([1, 3]),
            [DelayedError, ConflictError, "???" ])

    def testConflictResolutionTriggered2(self):
        """ Check that conflict resolution works """
        cluster = NEOCluster()
        cluster.start()
        try:
            # create the initial object
            t, c = cluster.getTransaction()
            c.root()['with_resolution'] = ob = PCounterWithResolution()
            t.commit()
            self.assertEqual(ob._p_changed, 0)
            oid = ob._p_oid
            tid1 = ob._p_serial
            self.assertNotEqual(tid1, ZERO_TID)
            del ob, t, c

            # then check resolution
            t1, c1 = cluster.getTransaction()
            t2, c2 = cluster.getTransaction()
            o1 = c1.root()['with_resolution']
            o2 = c2.root()['with_resolution']
            self.assertEqual(o1.value, 0)
            self.assertEqual(o2.value, 0)
            o1.value += 1
            o2.value += 2
            t1.commit()
            self.assertEqual(o1._p_changed, 0)
            tid2 = o1._p_serial
            self.assertTrue(tid1 < tid2)
            self.assertEqual(o1.value, 1)
            self.assertEqual(o2.value, 2)
            t2.commit()
            self.assertEqual(o2._p_changed, None)
            t1.begin()
            t2.begin()
            self.assertEqual(o2.value, 3)
            self.assertEqual(o1.value, 3)
            tid3 = o1._p_serial
            self.assertTrue(tid2 < tid3)
            self.assertEqual(tid3, o2._p_serial)

            # check history
            history = c1.db().history
            self.assertEqual([x['tid'] for x in history(oid, size=1)], [tid3])
            self.assertEqual([x['tid'] for x in history(oid, size=10)],
                             [tid3, tid2, tid1])
        finally:
            cluster.stop()

    def test_notifyNodeInformation(self):
        # translated from MasterNotificationsHandlerTests
        # (neo.tests.client.testMasterHandler)
        cluster = NEOCluster()
        try:
            cluster.start()
            cluster.db # open DB
            cluster.client.setPoll(0)
            storage, = cluster.client.nm.getStorageList()
            conn = storage.getConnection()
            self.assertFalse(conn.isClosed())
            getCellSortKey = cluster.client.cp.getCellSortKey
            self.assertEqual(getCellSortKey(storage), CELL_CONNECTED)
            cluster.neoctl.dropNode(cluster.storage.uuid)
            self.assertFalse(cluster.client.nm.getStorageList())
            self.assertTrue(conn.isClosed())
            self.assertEqual(getCellSortKey(storage), CELL_GOOD)
            # XXX: the test originally checked that 'unregister' method
            #      was called (even if it's useless in this case),
            #      but we would need an API to do that easily.
            self.assertFalse(cluster.client.dispatcher.registered(conn))
        finally:
            cluster.stop()

    def testRestartWithMissingStorage(self, fast_startup=False):
        # translated from neo.tests.functional.testStorage.StorageTest
        cluster = NEOCluster(replicas=1, partitions=10)
        s1, s2 = cluster.storage_list
        try:
            cluster.start()
            self.assertEqual([], cluster.getOudatedCells())
        finally:
            cluster.stop()
        # restart it with one storage only
        cluster.reset()
        try:
            cluster.start(storage_list=(s1,), fast_startup=fast_startup)
            self.assertEqual(NodeStates.UNKNOWN, cluster.getNodeState(s2))
        finally:
            cluster.stop()

    def testRestartWithMissingStorageFastStartup(self):
        self.testRestartWithMissingStorage(True)

    def testVerificationCommitUnfinishedTransactions(self, fast_startup=False):
        """ Verification step should commit unfinished transactions """
        # translated from neo.tests.functional.testCluster.ClusterTests
        cluster = NEOCluster()
        try:
            cluster.start()
            t, c = cluster.getTransaction()
            c.root()[0] = 'ok'
            t.commit()
        finally:
            cluster.stop()
        cluster.reset()
        # XXX: (obj|trans) become t(obj|trans)
        cluster.storage.switchTables()
        try:
            cluster.start(fast_startup=fast_startup)
            t, c = cluster.getTransaction()
            # transaction should be verified and commited
            self.assertEqual(c.root()[0], 'ok')
        finally:
            cluster.stop()

    def testVerificationCommitUnfinishedTransactionsFastStartup(self):
        self.testVerificationCommitUnfinishedTransactions(True)

    def testStorageReconnectDuringStore(self):
        cluster = NEOCluster(replicas=1)
        try:
            cluster.start()
            t, c = cluster.getTransaction()
            c.root()[0] = 'ok'
            while cluster.client.cp.connection_dict:
                cluster.client.cp._dropConnections()
            t.commit() # store request
        finally:
            cluster.stop()

    # The following 2 tests fail because the same queue is used for
    # AskTIDs(From) requests and reconnections. The same bug affected
    # history() before df47e5b1df8eabbff1383348b6b8c476bca0c328

    def testStorageReconnectDuringTransactionLog(self):
        cluster = NEOCluster(storage_count=2, partitions=2)
        try:
            cluster.start()
            t, c = cluster.getTransaction()
            while cluster.client.cp.connection_dict:
                cluster.client.cp._dropConnections()
            tid, (t1,) = cluster.client.transactionLog(
                ZERO_TID, c.root()._p_serial, 10)
        finally:
            cluster.stop()

    def testStorageReconnectDuringUndoLog(self):
        cluster = NEOCluster(storage_count=2, partitions=2)
        try:
            cluster.start()
            t, c = cluster.getTransaction()
            while cluster.client.cp.connection_dict:
                cluster.client.cp._dropConnections()
            t1, = cluster.client.undoLog(0, 10)
        finally:
            cluster.stop()



#
# Copyright (C) 2009-2010  Nexedi SA
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.

import unittest
from mock import Mock
from struct import pack, unpack
from neo.tests import NeoUnitTestBase
from neo.lib.protocol import NodeTypes, NodeStates, Packets
from neo.master.handlers.client import ClientServiceHandler
from neo.master.app import Application

class MasterClientHandlerTests(NeoUnitTestBase):

    def setUp(self):
        NeoUnitTestBase.setUp(self)
        # create an application object
        config = self.getMasterConfiguration(master_number=1, replicas=1)
        self.app = Application(config)
        self.app.pt.clear()
        self.app.pt.setID(1)
        self.app.em = Mock()
        self.app.loid = '\0' * 8
        self.app.tm.setLastTID('\0' * 8)
        self.service = ClientServiceHandler(self.app)
        # define some variable to simulate client and storage node
        self.client_port = 11022
        self.storage_port = 10021
        self.master_port = 10010
        self.master_address = ('127.0.0.1', self.master_port)
        self.client_address = ('127.0.0.1', self.client_port)
        self.storage_address = ('127.0.0.1', self.storage_port)
        # register the storage
        kw = {'uuid':self.getNewUUID(), 'address': self.master_address}
        self.app.nm.createStorage(**kw)

    def getLastUUID(self):
        return self.uuid

    def identifyToMasterNode(self, node_type=NodeTypes.STORAGE, ip="127.0.0.1",
                             port=10021):
        """Do first step of identification to MN """
        # register the master itself
        uuid = self.getNewUUID()
        self.app.nm.createFromNodeType(
            node_type,
            address=(ip, port),
            uuid=uuid,
            state=NodeStates.RUNNING,
        )
        return uuid

    # Tests
    def test_07_askBeginTransaction(self):
        tid1 = self.getNextTID()
        tid2 = self.getNextTID()
        service = self.service
        tm_org = self.app.tm
        self.app.tm = tm = Mock({
            'begin': '\x00\x00\x00\x00\x00\x00\x00\x01',
        })
        # client call it
        client_uuid = self.identifyToMasterNode(node_type=NodeTypes.CLIENT, port=self.client_port)
        client_node = self.app.nm.getByUUID(client_uuid)
        conn = self.getFakeConnection(client_uuid, self.client_address)
        service.askBeginTransaction(conn, None)
        calls = tm.mockGetNamedCalls('begin')
        self.assertEqual(len(calls), 1)
        calls[0].checkArgs(client_node, None)
        self.checkAnswerBeginTransaction(conn)
        # Client asks for a TID
        conn = self.getFakeConnection(client_uuid, self.client_address)
        self.app.tm = tm_org
        service.askBeginTransaction(conn, tid1)
        calls = tm.mockGetNamedCalls('begin')
        self.assertEqual(len(calls), 1)
        calls[0].checkArgs(client_node, None)
        args = self.checkAnswerBeginTransaction(conn, decode=True)
        self.assertEqual(args, (tid1, ))

    def test_08_askNewOIDs(self):
        service = self.service
        oid1, oid2 = self.getOID(1), self.getOID(2)
        self.app.tm.setLastOID(oid1)
        # client call it
        client_uuid = self.identifyToMasterNode(node_type=NodeTypes.CLIENT, port=self.client_port)
        conn = self.getFakeConnection(client_uuid, self.client_address)
        for node in self.app.nm.getStorageList():
            conn = self.getFakeConnection(node.getUUID(), node.getAddress())
            node.setConnection(conn)
        service.askNewOIDs(conn, 1)
        self.assertTrue(self.app.tm.getLastOID() > oid1)
        for node in self.app.nm.getStorageList():
            conn = node.getConnection()
            self.assertEqual(self.checkNotifyLastOID(conn, decode=True), (oid2,))

    def test_09_askFinishTransaction(self):
        service = self.service
        uuid = self.identifyToMasterNode()
        # do the right job
        client_uuid = self.identifyToMasterNode(node_type=NodeTypes.CLIENT, port=self.client_port)
        storage_uuid = self.identifyToMasterNode()
        storage_conn = self.getFakeConnection(storage_uuid, self.storage_address)
        storage2_uuid = self.identifyToMasterNode()
        storage2_conn = self.getFakeConnection(storage2_uuid,
            (self.storage_address[0], self.storage_address[1] + 1))
        self.app.setStorageReady(storage2_uuid)
        self.assertNotEqual(uuid, client_uuid)
        conn = self.getFakeConnection(client_uuid, self.client_address)
        self.app.pt = Mock({
            'getPartition': 0,
            'getCellList': [
                Mock({'getUUID': storage_uuid}),
                Mock({'getUUID': storage2_uuid}),
            ],
            'getPartitions': 2,
        })
        ttid = self.getNextTID()
        service.askBeginTransaction(conn, ttid)
        oid_list = []
        conn = self.getFakeConnection(client_uuid, self.client_address)
        self.app.nm.getByUUID(storage_uuid).setConnection(storage_conn)
        # No packet sent if storage node is not ready
        self.assertFalse(self.app.isStorageReady(storage_uuid))
        service.askFinishTransaction(conn, ttid, oid_list)
        self.checkNoPacketSent(storage_conn)
        self.app.tm.abortFor(self.app.nm.getByUUID(client_uuid))
        # ...but AskLockInformation is sent if it is ready
        self.app.setStorageReady(storage_uuid)
        self.assertTrue(self.app.isStorageReady(storage_uuid))
        service.askFinishTransaction(conn, ttid, oid_list)
        self.checkAskLockInformation(storage_conn)
        self.assertEqual(len(self.app.tm.registerForNotification(storage_uuid)), 1)
        txn = self.app.tm[ttid]
        pending_ttid = list(self.app.tm.registerForNotification(storage_uuid))[0]
        self.assertEqual(ttid, pending_ttid)
        self.assertEqual(len(txn.getOIDList()), 0)
        self.assertEqual(len(txn.getUUIDList()), 1)

    def test_askNodeInformations(self):
        # check that only informations about master and storages nodes are
        # send to a client
        self.app.nm.createClient()
        conn = self.getFakeConnection()
        self.service.askNodeInformation(conn)
        calls = conn.mockGetNamedCalls('notify')
        self.assertEqual(len(calls), 1)
        packet = calls[0].getParam(0)
        (node_list, ) = packet.decode()
        self.assertEqual(len(node_list), 2)

    def test_connectionClosed(self):
        # give a client uuid which have unfinished transactions
        client_uuid = self.identifyToMasterNode(node_type=NodeTypes.CLIENT,
                                                port = self.client_port)
        conn = self.getFakeConnection(client_uuid, self.client_address)
        self.app.listening_conn = object() # mark as running
        lptid = self.app.pt.getID()
        self.assertEqual(self.app.nm.getByUUID(client_uuid).getState(),
                NodeStates.RUNNING)
        self.service.connectionClosed(conn)
        # node must be have been remove, and no more transaction must remains
        self.assertEqual(self.app.nm.getByUUID(client_uuid), None)
        self.assertEqual(lptid, self.app.pt.getID())

    def test_askPack(self):
        self.assertEqual(self.app.packing, None)
        self.app.nm.createClient()
        tid = self.getNextTID()
        peer_id = 42
        conn = self.getFakeConnection(peer_id=peer_id)
        storage_uuid = self.identifyToMasterNode()
        storage_conn = self.getFakeConnection(storage_uuid,
            self.storage_address)
        self.app.nm.getByUUID(storage_uuid).setConnection(storage_conn)
        self.service.askPack(conn, tid)
        self.checkNoPacketSent(conn)
        ptid = self.checkAskPacket(storage_conn, Packets.AskPack,
            decode=True)[0]
        self.assertEqual(ptid, tid)
        self.assertTrue(self.app.packing[0] is conn)
        self.assertEqual(self.app.packing[1], peer_id)
        self.assertEqual(self.app.packing[2], set([storage_uuid, ]))
        # Asking again to pack will cause an immediate error
        storage_uuid = self.identifyToMasterNode()
        storage_conn = self.getFakeConnection(storage_uuid,
            self.storage_address)
        self.app.nm.getByUUID(storage_uuid).setConnection(storage_conn)
        self.service.askPack(conn, tid)
        self.checkNoPacketSent(storage_conn)
        status = self.checkAnswerPacket(conn, Packets.AnswerPack,
            decode=True)[0]
        self.assertFalse(status)

if __name__ == '__main__':
    unittest.main()




#
# Copyright (C) 2006-2010  Nexedi SA
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.

from time import time

import neo.lib
from neo.lib.util import dump
from neo.lib.protocol import NodeTypes, NodeStates

from neo.lib import attributeTracker

class Node(object):
    """This class represents a node."""

    _connection = None

    def __init__(self, manager, address=None, uuid=None,
            state=NodeStates.UNKNOWN):
        self._state = state
        self._address = address
        self._uuid = uuid
        self._manager = manager
        self._last_state_change = time()
        manager.add(self)

    def notify(self, packet):
        assert self.isConnected(), 'Not connected'
        self._connection.notify(packet)

    def ask(self, packet, *args, **kw):
        assert self.isConnected(), 'Not connected'
        self._connection.ask(packet, *args, **kw)

    def answer(self, packet, msg_id=None):
        assert self.isConnected(), 'Not connected'
        self._connection.answer(packet, msg_id)

    def getLastStateChange(self):
        return self._last_state_change

    def getState(self):
        return self._state

    def setState(self, new_state):
        if self._state == new_state:
            return
        old_state = self._state
        self._state = new_state
        self._last_state_change = time()
        self._manager._updateState(self, old_state)

    def setAddress(self, address):
        if self._address == address:
            return
        old_address = self._address
        self._address = address
        self._manager._updateAddress(self, old_address)

    def getAddress(self):
        return self._address

    def setUUID(self, uuid):
        if self._uuid == uuid:
            return
        old_uuid = self._uuid
        self._uuid = uuid
        self._manager._updateUUID(self, old_uuid)
        self._manager._updateIdentified(self)

    def getUUID(self):
        return self._uuid

    def onConnectionClosed(self):
        """
            Callback from node's connection when closed
        """
        assert self._connection is not None
        del self._connection
        self._manager._updateIdentified(self)

    def setConnection(self, connection):
        """
            Define the connection that is currently available to this node.
        """
        assert connection is not None
        assert self._connection is None
        self._connection = connection
        connection.setOnClose(self.onConnectionClosed)
        self._manager._updateIdentified(self)

    def getConnection(self):
        """
            Returns the connection to the node if available
        """
        assert self._connection is not None
        return self._connection

    def isConnected(self):
        """
            Returns True is a connection is established with the node
        """
        return self._connection is not None

    def isIdentified(self):
        """
            Returns True is the node is connected and identified
        """
        return self._connection is not None and self._uuid is not None

    def __repr__(self):
        return '<%s(uuid=%s, address=%s, state=%s) at %x>' % (
            self.__class__.__name__,
            dump(self._uuid),
            self._address,
            self._state,
            id(self),
        )

    def isMaster(self):
        return False

    def isStorage(self):
        return False

    def isClient(self):
        return False

    def isAdmin(self):
        return False

    def isRunning(self):
        return self._state == NodeStates.RUNNING

    def isUnknown(self):
        return self._state == NodeStates.UNKNOWN

    def isTemporarilyDown(self):
        return self._state == NodeStates.TEMPORARILY_DOWN

    def isDown(self):
        return self._state == NodeStates.DOWN

    def isBroken(self):
        return self._state == NodeStates.BROKEN

    def isHidden(self):
        return self._state == NodeStates.HIDDEN

    def isPending(self):
        return self._state == NodeStates.PENDING

    def setRunning(self):
        self.setState(NodeStates.RUNNING)

    def setUnknown(self):
        self.setState(NodeStates.UNKNOWN)

    def setTemporarilyDown(self):
        self.setState(NodeStates.TEMPORARILY_DOWN)

    def setDown(self):
        self.setState(NodeStates.DOWN)

    def setBroken(self):
        self.setState(NodeStates.BROKEN)

    def setHidden(self):
        self.setState(NodeStates.HIDDEN)

    def setPending(self):
        self.setState(NodeStates.PENDING)

    def asTuple(self):
        """ Returned tuple is intented to be used in procotol encoders """
        return (self.getType(), self._address, self._uuid, self._state)

    def __gt__(self, node):
        # sort per UUID if defined
        if self._uuid is not None:
            return self._uuid > node._uuid
        return self._address > node._address

    def getType(self):
        try:
            return NODE_CLASS_MAPPING[self.__class__]
        except KeyError:
            raise NotImplementedError

    def whoSetState(self):
        """
          Debugging method: call this method to know who set the current
          state value.
        """
        return attributeTracker.whoSet(self, '_state')

attributeTracker.track(Node)

class MasterNode(Node):
    """This class represents a master node."""

    def isMaster(self):
        return True

class StorageNode(Node):
    """This class represents a storage node."""

    def isStorage(self):
        return True

class ClientNode(Node):
    """This class represents a client node."""

    def isClient(self):
        return True

class AdminNode(Node):
    """This class represents an admin node."""

    def isAdmin(self):
        return True


NODE_TYPE_MAPPING = {
    NodeTypes.MASTER: MasterNode,
    NodeTypes.STORAGE: StorageNode,
    NodeTypes.CLIENT: ClientNode,
    NodeTypes.ADMIN: AdminNode,
}
NODE_CLASS_MAPPING = {
    StorageNode: NodeTypes.STORAGE,
    MasterNode: NodeTypes.MASTER,
    ClientNode: NodeTypes.CLIENT,
    AdminNode: NodeTypes.ADMIN,
}

class NodeManager(object):
    """This class manages node status."""

    # TODO: rework getXXXList() methods, filter first by node type
    # - getStorageList(identified=True, connected=True, )
    # - getList(...)

    def __init__(self):
        self._node_set = set()
        self._address_dict = {}
        self._uuid_dict = {}
        self._type_dict = {}
        self._state_dict = {}
        self._identified_dict = {}

    close = __init__

    def add(self, node):
        if node in self._node_set:
            neo.lib.logging.warning('adding a known node %r, ignoring', node)
            return
        self._node_set.add(node)
        self._updateAddress(node, None)
        self._updateUUID(node, None)
        self.__updateSet(self._type_dict, None, node.__class__, node)
        self.__updateSet(self._state_dict, None, node.getState(), node)
        self._updateIdentified(node)

    def remove(self, node):
        if node not in self._node_set:
            neo.lib.logging.warning('removing unknown node %r, ignoring', node)
            return
        self._node_set.remove(node)
        self.__drop(self._address_dict, node.getAddress())
        self.__drop(self._uuid_dict, node.getUUID())
        self.__dropSet(self._state_dict, node.getState(), node)
        self.__dropSet(self._type_dict, node.__class__, node)
        uuid = node.getUUID()
        if uuid in self._identified_dict:
            del self._identified_dict[uuid]

    def __drop(self, index_dict, key):
        try:
            del index_dict[key]
        except KeyError:
            # a node may have not be indexed by uuid or address, eg.:
            # - a master known by address but without UUID
            # - a client or admin node that don't have listening address
            pass

    def __update(self, index_dict, old_key, new_key, node):
        """ Update an index from old to new key """
        if old_key is not None:
            assert index_dict[old_key] is node, '%r is stored as %s, ' \
                'moving %r to %s' % (index_dict[old_key], old_key, node,
                new_key)
            del index_dict[old_key]
        if new_key is not None:
            index_dict[new_key] = node

    def _updateIdentified(self, node):
        uuid = node.getUUID()
        if node.isIdentified():
            self._identified_dict[uuid] = node
        else:
            self._identified_dict.pop(uuid, None)

    def _updateAddress(self, node, old_address):
        self.__update(self._address_dict, old_address, node.getAddress(), node)

    def _updateUUID(self, node, old_uuid):
        self.__update(self._uuid_dict, old_uuid, node.getUUID(), node)

    def __dropSet(self, set_dict, key, node):
        if key in set_dict and node in set_dict[key]:
            set_dict[key].remove(node)

    def __updateSet(self, set_dict, old_key, new_key, node):
        """ Update a set index from old to new key """
        if old_key in set_dict:
            set_dict[old_key].remove(node)
        if new_key is not None:
            set_dict.setdefault(new_key, set()).add(node)

    def _updateState(self, node, old_state):
        self.__updateSet(self._state_dict, old_state, node.getState(), node)

    def getList(self, node_filter=None):
        if filter is None:
            return list(self._node_set)
        return filter(node_filter, self._node_set)

    def getIdentifiedList(self, pool_set=None):
        """
            Returns a generator to iterate over identified nodes
            pool_set is an iterable of UUIDs allowed
        """
        if pool_set is not None:
            identified_nodes = self._identified_dict.items()
            return [v for k, v in identified_nodes if k in pool_set]
        return list(self._identified_dict.values())

    def getConnectedList(self):
        """
            Returns a generator to iterate over connected nodes
        """
        # TODO: use an index
        return [x for x in self._node_set if x.isConnected()]

    def __getList(self, index_dict, key):
        return index_dict.setdefault(key, set())

    def getByStateList(self, state):
        """ Get a node list filtered per the node state """
        return list(self.__getList(self._state_dict, state))

    def __getTypeList(self, type_klass, only_identified=False):
        node_set = self.__getList(self._type_dict, type_klass)
        if only_identified:
            return [x for x in node_set if x.getUUID() in self._identified_dict]
        return list(node_set)

    def getMasterList(self, only_identified=False):
        """ Return a list with master nodes """
        return self.__getTypeList(MasterNode, only_identified)

    def getStorageList(self, only_identified=False):
        """ Return a list with storage nodes """
        return self.__getTypeList(StorageNode, only_identified)

    def getClientList(self, only_identified=False):
        """ Return a list with client nodes """
        return self.__getTypeList(ClientNode, only_identified)

    def getAdminList(self, only_identified=False):
        """ Return a list with admin nodes """
        return self.__getTypeList(AdminNode, only_identified)

    def getByAddress(self, address):
        """ Return the node that match with a given address """
        return self._address_dict.get(address, None)

    def getByUUID(self, uuid):
        """ Return the node that match with a given UUID """
        return self._uuid_dict.get(uuid, None)

    def hasAddress(self, address):
        return address in self._address_dict

    def hasUUID(self, uuid):
        return uuid in self._uuid_dict

    def _createNode(self, klass, **kw):
        return klass(self, **kw)

    def createMaster(self, **kw):
        """ Create and register a new master """
        return self._createNode(MasterNode, **kw)

    def createStorage(self, **kw):
        """ Create and register a new storage """
        return self._createNode(StorageNode, **kw)

    def createClient(self, **kw):
        """ Create and register a new client """
        return self._createNode(ClientNode, **kw)

    def createAdmin(self, **kw):
        """ Create and register a new admin """
        return self._createNode(AdminNode, **kw)

    def _getClassFromNodeType(self, node_type):
        klass = NODE_TYPE_MAPPING.get(node_type)
        if klass is None:
            raise ValueError('Unknown node type : %s' % node_type)
        return klass

    def createFromNodeType(self, node_type, **kw):
        return self._createNode(self._getClassFromNodeType(node_type), **kw)

    def init(self):
        self._node_set.clear()
        self._type_dict.clear()
        self._state_dict.clear()
        self._uuid_dict.clear()
        self._address_dict.clear()

    def update(self, node_list):
        for node_type, addr, uuid, state in node_list:
            # This should be done here (although klass might not be used in this
            # iteration), as it raises if type is not valid.
            klass = self._getClassFromNodeType(node_type)

            # lookup in current table
            node_by_uuid = self.getByUUID(uuid)
            node_by_addr = self.getByAddress(addr)
            node = node_by_uuid or node_by_addr

            log_args = (node_type, dump(uuid), addr, state)
            if node is None:
                if state == NodeStates.DOWN:
                    neo.lib.logging.debug('NOT creating node %s %s %s %s',
                        *log_args)
                else:
                    node = self._createNode(klass, address=addr, uuid=uuid,
                            state=state)
                    neo.lib.logging.debug('creating node %r', node)
            else:
                assert isinstance(node, klass), 'node %r is not ' \
                    'of expected type: %r' % (node, klass)
                assert None in (node_by_uuid, node_by_addr) or \
                    node_by_uuid is node_by_addr, \
                    'Discrepancy between node_by_uuid (%r) and ' \
                    'node_by_addr (%r)' % (node_by_uuid, node_by_addr)
                if state == NodeStates.DOWN:
                    neo.lib.logging.debug(
                                    'droping node %r (%r), found with %s ' \
                        '%s %s %s', node, node.isConnected(), *log_args)
                    if node.isConnected():
                        # cut this connection, node removed by handler
                        node.getConnection().close()
                    self.remove(node)
                else:
                    neo.lib.logging.debug('updating node %r to %s %s %s %s',
                        node, *log_args)
                    node.setUUID(uuid)
                    node.setAddress(addr)
                    node.setState(state)
        self.log()

    def log(self):
        neo.lib.logging.info('Node manager : %d nodes' % len(self._node_set))
        for node in sorted(list(self._node_set)):
            uuid = dump(node.getUUID()) or '-' * 32
            address = node.getAddress() or ''
            if address:
                address = '%s:%d' % address
            neo.lib.logging.info(' * %32s | %8s | %22s | %s' % (
                uuid, node.getType(), address, node.getState()))




#
# Copyright (C) 2006-2010  Nexedi SA
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.

import neo.lib.python
import logging as logging_std

FMT = ('%(asctime)s %(levelname)-9s %(name)-10s'
       ' [%(module)14s:%(lineno)3d] \n%(message)s')

class Formatter(logging_std.Formatter):

    def formatTime(self, record, datefmt=None):
        return logging_std.Formatter.formatTime(self, record,
           '%Y-%m-%d %H:%M:%S') + '.%04d' % (record.msecs * 10)

    def format(self, record):
        lines = iter(logging_std.Formatter.format(self, record).splitlines())
        prefix = lines.next()
        return '\n'.join(prefix + line for line in lines)

def setupLog(name='NEO', filename=None, verbose=False):
    global logging
    if verbose:
        level = logging_std.DEBUG
    else:
        level = logging_std.INFO
    if logging is not None:
        for handler in logging.handlers:
            handler.acquire()
            try:
                handler.close()
            finally:
                handler.release()
        del logging.manager.loggerDict[logging.name]
    logging = logging_std.getLogger(name)
    for handler in logging.handlers[:]:
        logging.removeHandler(handler)
    if filename is None:
        handler = logging_std.StreamHandler()
    else:
        handler = logging_std.FileHandler(filename)
    handler.setFormatter(Formatter(FMT))
    logging.setLevel(level)
    logging.addHandler(handler)
    logging.propagate = 0

# Create default logger
logging = None
setupLog()




#
# Copyright (C) 2006-2010  Nexedi SA
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.

import neo.lib
from random import choice

from neo.storage.handlers import replication
from neo.lib.protocol import NodeTypes, NodeStates, Packets
from neo.lib.connection import ClientConnection
from neo.lib.util import dump

class Partition(object):
    """This class abstracts the state of a partition."""

    def __init__(self, offset, max_tid, ttid_list):
        # Possible optimization:
        #   _pending_ttid_list & _critical_tid can be shared amongst partitions
        #   created at the same time (cf Replicator.setUnfinishedTIDList).
        #   Replicator.transactionFinished would only have to iterate on these
        #   different sets, instead of all partitions.
        self._offset = offset
        self._pending_ttid_list = set(ttid_list)
        # pending upper bound
        self._critical_tid = max_tid

    def getOffset(self):
        return self._offset

    def getCriticalTID(self):
        return self._critical_tid

    def transactionFinished(self, ttid, max_tid):
        self._pending_ttid_list.remove(ttid)
        assert max_tid is not None
        # final upper bound
        self._critical_tid = max_tid

    def safe(self):
        return not self._pending_ttid_list

class Task(object):
    """
    A Task is a callable to execute at another time, with given parameters.
    Execution result is kept and can be retrieved later.
    """

    _func = None
    _args = None
    _kw = None
    _result = None
    _processed = False

    def __init__(self, func, args=(), kw=None):
        self._func = func
        self._args = args
        if kw is None:
            kw = {}
        self._kw = kw

    def process(self):
        if self._processed:
            raise ValueError, 'You cannot process a single Task twice'
        self._processed = True
        self._result = self._func(*self._args, **self._kw)

    def getResult(self):
        # Should we instead execute immediately rather than raising ?
        if not self._processed:
            raise ValueError, 'You cannot get a result until task is executed'
        return self._result

    def __repr__(self):
        fmt = '<%s at %x %r(*%r, **%r)%%s>' % (self.__class__.__name__,
            id(self), self._func, self._args, self._kw)
        if self._processed:
            extra = ' => %r' % (self._result, )
        else:
            extra = ''
        return fmt % (extra, )

class Replicator(object):
    """This class handles replications of objects and transactions.

    Assumptions:

        - Client nodes recognize partition changes reasonably quickly.

        - When an out of date partition is added, next transaction ID
          is given after the change is notified and serialized.

    Procedures:

        - Get the last TID right after a partition is added. This TID
          is called a "critical TID", because this and TIDs before this
          may not be present in this storage node yet. After a critical
          TID, all transactions must exist in this storage node.

        - Check if a primary master node still has pending transactions
          before and at a critical TID. If so, I must wait for them to be
          committed or aborted.

        - In order to copy data, first get the list of TIDs. This is done
          part by part, because the list can be very huge. When getting
          a part of the list, I verify if they are in my database, and
          ask data only for non-existing TIDs. This is performed until
          the check reaches a critical TID.

        - Next, get the list of OIDs. And, for each OID, ask the history,
          namely, a list of serials. This is also done part by part, and
          I ask only non-existing data. """

    # new_partition_set
    #   outdated partitions for which no pending transactions was asked to
    #   primary master yet
    # partition_dict
    #   outdated partitions with pending transaction and temporary critical
    #   tid
    # current_partition
    #   partition being currently synchronised
    # current_connection
    #   connection to a storage node we are replicating from
    # waiting_for_unfinished_tids
    #   unfinished tids have been asked to primary master node, but it
    #   didn't answer yet.
    # replication_done
    #   False if we know there is something to replicate.
    #   True when current_partition is replicated, or we don't know yet if
    #   there is something to replicate

    current_partition = None
    current_connection = None
    waiting_for_unfinished_tids = False
    replication_done = True

    def __init__(self, app):
        self.app = app
        self.new_partition_set = set()
        self.partition_dict = {}
        self.task_list = []
        self.task_dict = {}

    def masterLost(self):
        """
        When connection to primary master is lost, stop waiting for unfinished
        transactions.
        """
        self.waiting_for_unfinished_tids = False

    def storageLost(self):
        """
        Restart replicating.
        """
        self.reset()

    def populate(self):
        """
        Populate partitions to replicate. Must be called when partition
        table is the one accepted by primary master.
        Implies a reset.
        """
        partition_list = self.app.pt.getOutdatedOffsetListFor(self.app.uuid)
        self.new_partition_set = set(partition_list)
        self.partition_dict = {}
        self.reset()

    def reset(self):
        """Reset attributes to restart replicating."""
        self.task_list = []
        self.task_dict = {}
        self.current_partition = None
        self.current_connection = None
        self.replication_done = True

    def pending(self):
        """Return whether there is any pending partition."""
        return bool(self.partition_dict or self.new_partition_set)

    def getCurrentOffset(self):
        assert self.current_partition is not None
        return self.current_partition.getOffset()

    def getCurrentCriticalTID(self):
        assert self.current_partition is not None
        return self.current_partition.getCriticalTID()

    def setReplicationDone(self):
        """ Callback from ReplicationHandler """
        self.replication_done = True

    def isCurrentConnection(self, conn):
        return self.current_connection is conn

    def setUnfinishedTIDList(self, max_tid, ttid_list):
        """This is a callback from MasterOperationHandler."""
        neo.lib.logging.debug('setting unfinished TTIDs %s',
                      ','.join([dump(tid) for tid in ttid_list]))
        # all new outdated partition must wait those ttid
        new_partition_set = self.new_partition_set
        while new_partition_set:
            offset = new_partition_set.pop()
            self.partition_dict[offset] = Partition(offset, max_tid, ttid_list)
        self.waiting_for_unfinished_tids = False

    def transactionFinished(self, ttid, max_tid):
        """ Callback from MasterOperationHandler """
        for partition in self.partition_dict.itervalues():
            partition.transactionFinished(ttid, max_tid)

    def _askUnfinishedTIDs(self):
        conn = self.app.master_conn
        conn.ask(Packets.AskUnfinishedTransactions())
        self.waiting_for_unfinished_tids = True

    def _startReplication(self):
        # Choose a storage node for the source.
        app = self.app
        cell_list = app.pt.getCellList(self.current_partition.getOffset(),
                                       readable=True)
        node_list = [cell.getNode() for cell in cell_list
                        if cell.getNodeState() == NodeStates.RUNNING]
        try:
            node = choice(node_list)
        except IndexError:
            # Not operational.
            neo.lib.logging.error('not operational', exc_info = 1)
            self.current_partition = None
            return

        addr = node.getAddress()
        if addr is None:
            neo.lib.logging.error("no address known for the selected node %s" %
                    (dump(node.getUUID()), ))
            return

        connection = self.current_connection
        if connection is None or connection.getAddress() != addr:
            handler = replication.ReplicationHandler(app)
            self.current_connection = ClientConnection(app.em, handler,
                   addr=addr, connector=app.connector_handler())
            p = Packets.RequestIdentification(NodeTypes.STORAGE,
                    app.uuid, app.server, app.name)
            self.current_connection.ask(p)
            if connection is not None:
                connection.close()
        else:
            connection.getHandler().startReplication(connection)
        self.replication_done = False

    def _finishReplication(self):
        # TODO: remove try..except: pass
        try:
            # Notify to a primary master node that my cell is now up-to-date.
            conn = self.app.master_conn
            offset = self.current_partition.getOffset()
            self.partition_dict.pop(offset)
            conn.notify(Packets.NotifyReplicationDone(offset))
        except KeyError:
            pass
        if self.pending():
            self.current_partition = None
        else:
            self.current_connection.close()

    def act(self):

        if self.current_partition is not None:
            # Don't end replication until we have received all expected
            # answers, as we might have asked object data just before the last
            # AnswerCheckSerialRange.
            if self.replication_done and \
                    not self.current_connection.isPending():
                # finish a replication
                neo.lib.logging.info('replication is done for %s' %
                        (self.current_partition.getOffset(), ))
                self._finishReplication()
            return

        if self.waiting_for_unfinished_tids:
            # Still waiting.
            neo.lib.logging.debug('waiting for unfinished tids')
            return

        if self.new_partition_set:
            # Ask pending transactions.
            neo.lib.logging.debug('asking unfinished tids')
            self._askUnfinishedTIDs()
            return

        # Try to select something.
        for partition in self.partition_dict.values():
            # XXX: replication could start up to the initial critical tid, that
            # is below the pending transactions, then finish when all pending
            # transactions are committed.
            if partition.safe():
                self.current_partition = partition
                break
        else:
            # Not yet.
            neo.lib.logging.debug('not ready yet')
            return

        self._startReplication()

    def removePartition(self, offset):
        """This is a callback from MasterOperationHandler."""
        self.partition_dict.pop(offset, None)
        self.new_partition_set.discard(offset)

    def addPartition(self, offset):
        """This is a callback from MasterOperationHandler."""
        if not self.partition_dict.has_key(offset):
            self.new_partition_set.add(offset)

    def _addTask(self, key, func, args=(), kw=None):
        task = Task(func, args, kw)
        task_dict = self.task_dict
        if key in task_dict:
            raise ValueError, 'Task with key %r already exists (%r), cannot ' \
                'add %r' % (key, task_dict[key], task)
        task_dict[key] = task
        self.task_list.append(task)

    def processDelayedTasks(self):
        task_list = self.task_list
        if task_list:
            for task in task_list:
                task.process()
            self.task_list = []

    def checkTIDRange(self, min_tid, max_tid, length, partition):
        app = self.app
        self._addTask(('TID', min_tid, length), app.dm.checkTIDRange,
            (min_tid, max_tid, length, app.pt.getPartitions(), partition))

    def checkSerialRange(self, min_oid, min_serial, max_tid, length,
            partition):
        app = self.app
        self._addTask(('Serial', min_oid, min_serial, length),
            app.dm.checkSerialRange, (min_oid, min_serial, max_tid, length,
            app.pt.getPartitions(), partition))

    def getTIDsFrom(self, min_tid, max_tid, length, partition):
        app = self.app
        self._addTask('TIDsFrom',
            app.dm.getReplicationTIDList, (min_tid, max_tid, length,
            app.pt.getPartitions(), partition))

    def getObjectHistoryFrom(self, min_oid, min_serial, max_serial, length,
            partition):
        app = self.app
        self._addTask('ObjectHistoryFrom',
            app.dm.getObjectHistoryFrom, (min_oid, min_serial, max_serial,
            length, app.pt.getPartitions(), partition))

    def _getCheckResult(self, key):
        return self.task_dict.pop(key).getResult()

    def getTIDCheckResult(self, min_tid, length):
        return self._getCheckResult(('TID', min_tid, length))

    def getSerialCheckResult(self, min_oid, min_serial, length):
        return self._getCheckResult(('Serial', min_oid, min_serial, length))

    def getTIDsFromResult(self):
        return self._getCheckResult('TIDsFrom')

    def getObjectHistoryFromResult(self):
        return self._getCheckResult('ObjectHistoryFrom')




#
# Copyright (C) 2009-2010  Nexedi SA
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.

import unittest
from mock import Mock
from neo.tests import NeoUnitTestBase
from neo.lib.protocol import NodeTypes, NotReadyError, \
        BrokenNodeDisallowedError
from neo.lib.pt import PartitionTable
from neo.storage.app import Application
from neo.storage.handlers.identification import IdentificationHandler

class StorageIdentificationHandlerTests(NeoUnitTestBase):

    def setUp(self):
        NeoUnitTestBase.setUp(self)
        config = self.getStorageConfiguration(master_number=1)
        self.app = Application(config)
        self.app.name = 'NEO'
        self.app.ready = True
        self.app.pt = PartitionTable(4, 1)
        self.identification = IdentificationHandler(self.app)

    def tearDown(self):
        self.app.close()
        del self.app
        super(StorageIdentificationHandlerTests, self).tearDown()

    def test_requestIdentification1(self):
        """ nodes are rejected during election or if unknown storage """
        self.app.ready = False
        self.assertRaises(
                NotReadyError,
                self.identification.requestIdentification,
                self.getFakeConnection(),
                NodeTypes.CLIENT,
                self.getNewUUID(),
                None,
                self.app.name,
        )
        self.app.ready = True
        self.assertRaises(
                NotReadyError,
                self.identification.requestIdentification,
                self.getFakeConnection(),
                NodeTypes.STORAGE,
                self.getNewUUID(),
                None,
                self.app.name,
        )

    def test_requestIdentification3(self):
        """ broken nodes must be rejected """
        uuid = self.getNewUUID()
        conn = self.getFakeConnection(uuid=uuid)
        node = self.app.nm.createClient(uuid=uuid)
        node.setBroken()
        self.assertRaises(BrokenNodeDisallowedError,
                self.identification.requestIdentification,
                conn,
                NodeTypes.CLIENT,
                uuid,
                None,
                self.app.name,
        )

    def test_requestIdentification2(self):
        """ accepted client must be connected and running """
        uuid = self.getNewUUID()
        conn = self.getFakeConnection(uuid=uuid)
        node = self.app.nm.createClient(uuid=uuid)
        self.identification.requestIdentification(conn, NodeTypes.CLIENT, uuid,
                None, self.app.name)
        self.assertTrue(node.isRunning())
        self.assertTrue(node.isConnected())
        self.assertEqual(node.getUUID(), uuid)
        self.assertTrue(node.getConnection() is conn)
        self.checkUUIDSet(conn, uuid)
        args = self.checkAcceptIdentification(conn, decode=True)
        node_type, address, _np, _nr, _uuid = args
        self.assertEqual(node_type, NodeTypes.STORAGE)
        self.assertEqual(address, None)
        self.assertEqual(_uuid, uuid)


if __name__ == "__main__":
    unittest.main()



##############################################################################
#
# Copyright (c) 2001, 2002 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE
#
##############################################################################

# At the moment, no ZODB release include the following patches.
# Later, this must be replaced by some detection mechanism.
if 1:

    from ZODB.Connection import Connection

    # Allow serial to be returned as late as tpc_finish
    #
    # This makes possible for storage to allocate serial inside tpc_finish,
    # removing the requirement to serialise second commit phase (tpc_vote
    # to tpc_finish/tpc_abort).

    def tpc_finish(self, transaction):
        """Indicate confirmation that the transaction is done."""

        def callback(tid):
            # BBB: _mvcc_storage not supported on older ZODB
            if getattr(self, '_mvcc_storage', False):
                # Inter-connection invalidation is not needed when the
                # storage provides MVCC.
                return
            d = dict.fromkeys(self._modified)
            self._db.invalidate(tid, d, self)
#       It's important that the storage calls the passed function
#       while it still has its lock.  We don't want another thread
#       to be able to read any updated data until we've had a chance
#       to send an invalidation message to all of the other
#       connections!
        serial = self._storage.tpc_finish(transaction, callback)
        if serial is not None:
            assert isinstance(serial, str), repr(serial)
            for oid_iterator in (self._modified, self._creating):
                for oid in oid_iterator:
                    obj = self._cache.get(oid, None)
                    # Ignore missing objects and don't update ghosts.
                    if obj is not None and obj._p_changed is not None:
                        obj._p_changed = 0
                        obj._p_serial = serial
        self._tpc_cleanup()

    Connection.tpc_finish = tpc_finish

    ###

    try:
        if Connection._nexedi_fix != 4:
            raise Exception("A different ZODB fix is already applied")
    except AttributeError:
        Connection._nexedi_fix = 4

        # Whenever an connection is opened (and there's usually an existing one
        # in DB pool that can be reused) whereas the transaction is already
        # started, we must make sure that proper storage setup is done by
        # calling Connection.newTransaction.
        # For example, there's no open transaction when a ZPublisher/Publish
        # transaction begins.

        import thread

        def open(self, *args, **kw):
            def _flush_invalidations():
                acquire = self._db._a
                try:
                    self._db._r()
                except thread.error:
                    acquire = lambda: None
                try:
                    del self._flush_invalidations
                    self.newTransaction()
                finally:
                    acquire()
                    self._flush_invalidations = _flush_invalidations
            self._flush_invalidations = _flush_invalidations
            try:
                Connection_open(self, *args, **kw)
            finally:
                del self._flush_invalidations
        try:
            Connection_open = Connection._setDB
            Connection._setDB = open
        except AttributeError: # recent ZODB
            Connection_open = Connection.open
            Connection.open = open

        # Storage.sync usually implements a "network barrier" (at least
        # in NEO, but ZEO should be fixed to do the same), which is quite
        # slow so we prefer to not call it where it's not useful.
        # I don't know any legitimate use of DB access outside a transaction.

        # But old versions of ERP5 (before 2010-10-29 17:15:34) and maybe other
        # applications do not always call 'transaction.begin()' when they should
        # so this patch disabled as a precaution, at least as long as we support
        # old software. This should also be discussed on zodb-dev ML first.

        def afterCompletion(self, *ignored):
            try:
                self._readCurrent.clear()
            except AttributeError: # old ZODB (e.g. ZODB 3.4)
                pass
            self._flush_invalidations()
        #Connection.afterCompletion = afterCompletion


    class _DB(object):
        """
        Wrapper to DB instance that properly initialize Connection objects
        with NEO storages.
        It forces the connection to always create a new instance of the
        storage, for compatibility with ZODB 3.4, and because we don't
        implement IMVCCStorage completely.
        """

        def __new__(cls, db, connection):
            if db._storage.__class__.__module__ != 'neo.client.Storage':
                return db
            self = object.__new__(cls)
            self._db = db
            self._connection = connection
            return self

        def __getattr__(self, attr):
            result = getattr(self._db, attr)
            if attr in ('storage', '_storage'):
                result = result.new_instance()
                self._connection._db = self._db
                setattr(self, attr, result)
            return result

    try:
        Connection_setDB = Connection._setDB
    except AttributeError: # recent ZODB
        Connection_init = Connection.__init__
        Connection.__init__ = lambda self, db, *args, **kw: \
            Connection_init(self, _DB(db, self), *args, **kw)
    else: # old ZODB (e.g. ZODB 3.4)
        Connection._setDB = lambda self, odb, *args, **kw: \
            Connection_setDB(self, _DB(odb, self), *args, **kw)

        from ZODB.DB import DB
        DB_invalidate = DB.invalidate
        DB.invalidate = lambda self, tid, oids, *args, **kw: \
            DB_invalidate(self, tid, dict.fromkeys(oids, None), *args, **kw)



#
# Copyright (C) 2006-2010  Nexedi SA
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.

from neo.lib import util
from neo.lib.exception import DatabaseFailure

class CreationUndone(Exception):
    pass

class DatabaseManager(object):
    """This class only describes an interface for database managers."""

    def __init__(self):
        """
            Initialize the object.
        """
        self._under_transaction = False

    def isUnderTransaction(self):
        return self._under_transaction

    def begin(self):
        """
            Begin a transaction
        """
        if self._under_transaction:
            raise DatabaseFailure('A transaction has already begun')
        self._begin()
        self._under_transaction = True

    def commit(self):
        """
            Commit the current transaction
        """
        if not self._under_transaction:
            raise DatabaseFailure('The transaction has not begun')
        self._commit()
        self._under_transaction = False

    def rollback(self):
        """
            Rollback the current transaction
        """
        self._rollback()
        self._under_transaction = False

    def setup(self, reset = 0):
        """Set up a database. If reset is true, existing data must be
        discarded."""
        raise NotImplementedError

    def _begin(self):
        raise NotImplementedError

    def _commit(self):
        raise NotImplementedError

    def _rollback(self):
        raise NotImplementedError

    def _getPartition(self, oid_or_tid):
        return oid_or_tid % self.getNumPartitions()

    def getConfiguration(self, key):
        """
            Return a configuration value, returns None if not found or not set
        """
        raise NotImplementedError

    def setConfiguration(self, key, value):
        """
            Set a configuration value
        """
        if self._under_transaction:
            self._setConfiguration(key, value)
        else:
            self.begin()
            try:
                self._setConfiguration(key, value)
            except:
                self.rollback()
                raise
            self.commit()

    def _setConfiguration(self, key, value):
        raise NotImplementedError

    def getUUID(self):
        """
            Load an UUID from a database.
        """
        return util.bin(self.getConfiguration('uuid'))

    def setUUID(self, uuid):
        """
            Store an UUID into a database.
        """
        self.setConfiguration('uuid', util.dump(uuid))

    def getNumPartitions(self):
        """
            Load the number of partitions from a database.
        """
        n = self.getConfiguration('partitions')
        if n is not None:
            return int(n)

    def setNumPartitions(self, num_partitions):
        """
            Store the number of partitions into a database.
        """
        self.setConfiguration('partitions', num_partitions)

    def getNumReplicas(self):
        """
            Load the number of replicas from a database.
        """
        n = self.getConfiguration('replicas')
        if n is not None:
            return int(n)

    def setNumReplicas(self, num_replicas):
        """
            Store the number of replicas into a database.
        """
        self.setConfiguration('replicas', num_replicas)

    def getName(self):
        """
            Load a name from a database.
        """
        return self.getConfiguration('name')

    def setName(self, name):
        """
            Store a name into a database.
        """
        self.setConfiguration('name', name)

    def getPTID(self):
        """
            Load a Partition Table ID from a database.
        """
        return long(self.getConfiguration('ptid'))

    def setPTID(self, ptid):
        """
            Store a Partition Table ID into a database.
        """
        if ptid is not None:
            assert isinstance(ptid, (int, long)), ptid
            ptid = str(ptid)
        self.setConfiguration('ptid', ptid)

    def getLastOID(self):
        """
            Returns the last OID used
        """
        return util.bin(self.getConfiguration('loid'))

    def setLastOID(self, loid):
        """
            Set the last OID used
        """
        self.setConfiguration('loid', util.dump(loid))

    def getPartitionTable(self):
        """Return a whole partition table as a tuple of rows. Each row
        is again a tuple of an offset (row ID), an UUID of a storage
        node, and a cell state."""
        raise NotImplementedError

    def getLastTID(self, all = True):
        """Return the last TID in a database. If all is true,
        unfinished transactions must be taken account into. If there
        is no TID in the database, return None."""
        raise NotImplementedError

    def getUnfinishedTIDList(self):
        """Return a list of unfinished transaction's IDs."""
        raise NotImplementedError

    def objectPresent(self, oid, tid, all = True):
        """Return true iff an object specified by a given pair of an
        object ID and a transaction ID is present in a database.
        Otherwise, return false. If all is true, the object must be
        searched from unfinished transactions as well."""
        raise NotImplementedError

    def _getObject(self, oid, tid=None, before_tid=None):
        """
        oid (int)
            Identifier of object to retrieve.
        tid (int, None)
            Exact serial to retrieve.
        before_tid (packed, None)
            Serial to retrieve is the highest existing one strictly below this
            value.
        """
        raise NotImplementedError

    def getObject(self, oid, tid=None, before_tid=None, resolve_data=True):
        """
        oid (packed)
            Identifier of object to retrieve.
        tid (packed, None)
            Exact serial to retrieve.
        before_tid (packed, None)
            Serial to retrieve is the highest existing one strictly below this
            value.
        resolve_data (bool, True)
            If actual object data is desired, or raw record content.
            This is different in case retrieved line undoes a transaction.

        Return value:
            None: Given oid doesn't exist in database.
            False: No record found, but another one exists for given oid.
            6-tuple: Record content.
                - record serial (packed)
                - serial or next record modifying object (packed, None)
                - compression (boolean-ish, None)
                - checksum (integer, None)
                - data (binary string, None)
                - data_serial (packed, None)
        """
        # TODO: resolve_data must be unit-tested
        u64 = util.u64
        p64 = util.p64
        oid = u64(oid)
        if tid is not None:
            tid = u64(tid)
        if before_tid is not None:
            before_tid = u64(before_tid)
        result = self._getObject(oid, tid, before_tid)
        if result is None:
            # See if object exists at all
            result = self._getObject(oid)
            if result is not None:
                # Object exists
                result = False
        else:
            serial, next_serial, compression, checksum, data, data_serial = \
                result
            assert before_tid is None or next_serial is None or \
                   before_tid <= next_serial
            if data is None and resolve_data:
                try:
                    _, compression, checksum, data = self._getObjectData(oid,
                        data_serial, serial)
                except CreationUndone:
                    compression = 0
                    # XXX: this is the valid checksum for empty string
                    checksum = 1
                    data = ''
                data_serial = None
            if serial is not None:
                serial = p64(serial)
            if next_serial is not None:
                next_serial = p64(next_serial)
            if data_serial is not None:
                data_serial = p64(data_serial)
            result = serial, next_serial, compression, checksum, data, data_serial
        return result

    def changePartitionTable(self, ptid, cell_list):
        """Change a part of a partition table. The list of cells is
        a tuple of tuples, each of which consists of an offset (row ID),
        an UUID of a storage node, and a cell state. The Partition
        Table ID must be stored as well."""
        raise NotImplementedError

    def setPartitionTable(self, ptid, cell_list):
        """Set a whole partition table. The semantics is the same as
        changePartitionTable, except that existing data must be
        thrown away."""
        raise NotImplementedError

    def dropPartitions(self, num_partitions, offset_list):
        """ Drop any data of non-assigned partitions for a given UUID """
        raise NotImplementedError('this method must be overriden')

    def dropUnfinishedData(self):
        """Drop any unfinished data from a database."""
        raise NotImplementedError

    def storeTransaction(self, tid, object_list, transaction, temporary = True):
        """Store a transaction temporarily, if temporary is true. Note
        that this transaction is not finished yet. The list of objects
        contains tuples, each of which consists of an object ID,
        a compression specification, a checksum and object data.
        The transaction is either None or a tuple of the list of OIDs,
        user information, a description, extension information and transaction
        pack state (True for packed)."""
        raise NotImplementedError

    def _getDataTID(self, oid, tid=None, before_tid=None):
        """
        Return a 2-tuple:
        tid (int)
            tid corresponding to received parameters
        serial
            tid at which actual object data is located

        If 'tid is None', requested object and transaction could
        not be found.
        If 'serial is None', requested object exist but has no data (its creation
        has been undone).
        If 'tid == serial', it means that requested transaction
        contains object data.
        Otherwise, it's an undo transaction which did not involve conflict
        resolution.
        """
        raise NotImplementedError

    def findUndoTID(self, oid, tid, ltid, undone_tid, transaction_object):
        """
        oid
            Object OID
        tid
            Transation doing the undo
        ltid
            Upper (exclued) bound of transactions visible to transaction doing
            the undo.
        undone_tid
            Transaction to undo
        transaction_object
            Object data from memory, if it was modified by running
            transaction.
            None if is was not modified by running transaction.

        Returns a 3-tuple:
        current_tid (p64)
            TID of most recent version of the object client's transaction can
            see. This is used later to detect current conflicts (eg, another
            client modifying the same object in parallel)
        data_tid (int)
            TID containing (without indirection) the data prior to undone
            transaction.
            None if object doesn't exist prior to transaction being undone
              (its creation is being undone).
        is_current (bool)
            False if object was modified by later transaction (ie, data_tid is
            not current), True otherwise.
        """
        u64 = util.u64
        p64 = util.p64
        oid = u64(oid)
        tid = u64(tid)
        if ltid:
            ltid = u64(ltid)
        undone_tid = u64(undone_tid)
        _getDataTID = self._getDataTID
        if transaction_object is not None:
            _, _, _, _, tvalue_serial = transaction_object
            current_tid = current_data_tid = u64(tvalue_serial)
        else:
            current_tid, current_data_tid = _getDataTID(oid, before_tid=ltid)
        if current_tid is None:
            return (None, None, False)
        found_undone_tid, undone_data_tid = _getDataTID(oid, tid=undone_tid)
        assert found_undone_tid is not None, (oid, undone_tid)
        is_current = undone_data_tid in (current_data_tid, tid)
        # Load object data as it was before given transaction.
        # It can be None, in which case it means we are undoing object
        # creation.
        _, data_tid = _getDataTID(oid, before_tid=undone_tid)
        if data_tid is not None:
            data_tid = p64(data_tid)
        return p64(current_tid), data_tid, is_current

    def finishTransaction(self, tid):
        """Finish a transaction specified by a given ID, by moving
        temporarily data to a finished area."""
        raise NotImplementedError

    def deleteTransaction(self, tid, oid_list=()):
        """Delete a transaction and its content specified by a given ID and
        an oid list"""
        raise NotImplementedError

    def deleteTransactionsAbove(self, num_partitions, partition, tid, max_tid):
        """Delete all transactions above given TID (inclued) in given
        partition, but never above max_tid (in case transactions are committed
        during replication)."""
        raise NotImplementedError

    def deleteObject(self, oid, serial=None):
        """Delete given object. If serial is given, only delete that serial for
        given oid."""
        raise NotImplementedError

    def deleteObjectsAbove(self, num_partitions, partition, oid, serial,
            max_tid):
        """Delete all objects above given OID and serial (inclued) in given
        partition, but never above max_tid (in case objects are stored during
        replication)"""
        raise NotImplementedError

    def getTransaction(self, tid, all = False):
        """Return a tuple of the list of OIDs, user information,
        a description, and extension information, for a given transaction
        ID. If there is no such transaction ID in a database, return None.
        If all is true, the transaction must be searched from a temporary
        area as well."""
        raise NotImplementedError

    def getObjectHistory(self, oid, offset = 0, length = 1):
        """Return a list of serials and sizes for a given object ID.
        The length specifies the maximum size of such a list. Result starts
        with latest serial, and the list must be sorted in descending order.
        If there is no such object ID in a database, return None."""
        raise NotImplementedError

    def getObjectHistoryFrom(self, oid, min_serial, max_serial, length,
            num_partitions, partition):
        """Return a dict of length serials grouped by oid at (or above)
        min_oid and min_serial and below max_serial, for given partition,
        sorted in ascending order."""
        raise NotImplementedError

    def getTIDList(self, offset, length, num_partitions, partition_list):
        """Return a list of TIDs in ascending order from an offset,
        at most the specified length. The list of partitions are passed
        to filter out non-applicable TIDs."""
        raise NotImplementedError

    def getReplicationTIDList(self, min_tid, max_tid, length, num_partitions,
        partition):
        """Return a list of TIDs in ascending order from an initial tid value,
        at most the specified length up to max_tid. The partition number is
        passed to filter out non-applicable TIDs."""
        raise NotImplementedError

    def pack(self, tid, updateObjectDataForPack):
        """Prune all non-current object revisions at given tid.
        updateObjectDataForPack is a function called for each deleted object
        and revision with:
        - OID
        - packed TID
        - new value_serial
            If object data was moved to an after-pack-tid revision, this
            parameter contains the TID of that revision, allowing to backlink
            to it.
        - getObjectData function
            To call if value_serial is None and an object needs to be updated.
            Takes no parameter, returns a 3-tuple: compression, checksum,
            value
        """
        raise NotImplementedError

    def checkTIDRange(self, min_tid, max_tid, length, num_partitions, partition):
        """
        Generate a diggest from transaction list.
        min_tid (packed)
            TID at which verification starts.
        length (int)
            Maximum number of records to include in result.
        num_partitions, partition (int, int)
            Specifies concerned partition.

        Returns a 3-tuple:
            - number of records actually found
            - a XOR computed from record's TID
              0 if no record found
            - biggest TID found (ie, TID of last record read)
              ZERO_TID if not record found
        """
        raise NotImplementedError

    def checkSerialRange(self, min_oid, min_serial, max_tid, length,
            num_partitions, partition):
        """
        Generate a diggest from object list.
        min_oid (packed)
            OID at which verification starts.
        min_serial (packed)
            Serial of min_oid object at which search should start.
        length
            Maximum number of records to include in result.
        num_partitions, partition (int, int)
            Specifies concerned partition.

        Returns a 5-tuple:
            - number of records actually found
            - a XOR computed from record's OID
              0 if no record found
            - biggest OID found (ie, OID of last record read)
              ZERO_OID if no record found
            - a XOR computed from record's serial
              0 if no record found
            - biggest serial found for biggest OID found (ie, serial of last
              record read)
              ZERO_TID if no record found
        """
        raise NotImplementedError




#
# Copyright (C) 2006-2010  Nexedi SA
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.

from neo.neoctl.neoctl import NeoCTL, NotReadyException
from neo.lib.util import bin, dump
from neo.lib.protocol import ClusterStates, NodeStates, NodeTypes

action_dict = {
    'print': {
        'pt': 'getPartitionRowList',
        'node': 'getNodeList',
        'cluster': 'getClusterState',
        'primary': 'getPrimary',
    },
    'set': {
        'node': 'setNodeState',
        'cluster': 'setClusterState',
    },
    'start': 'startCluster',
    'add': 'enableStorageList',
    'drop': 'dropNode',
}

class TerminalNeoCTL(object):
    def __init__(self, address):
        self.neoctl = NeoCTL(address)

    def __del__(self):
        self.neoctl.close()

    # Utility methods (could be functions)
    def asNodeState(self, value):
        return NodeStates.getByName(value.upper())

    def asNodeType(self, value):
        return NodeTypes.getByName(value.upper())

    def asClusterState(self, value):
        return ClusterStates.getByName(value.upper())

    def asNode(self, value):
        return bin(value)

    def formatRowList(self, row_list):
        return '\n'.join('%03d | %s' % (offset,
            ''.join('%s - %s |' % (dump(uuid), state)
            for (uuid, state) in cell_list))
            for (offset, cell_list) in row_list)

    def formatNodeList(self, node_list):
        if not node_list:
            return 'Empty list!'
        result = []
        for node_type, address, uuid, state in node_list:
            if address is None:
                address = (None, None)
            ip, port = address
            result.append('%s - %s - %s:%s - %s' % (node_type, dump(uuid), ip,
                                                    port, state))
        return '\n'.join(result)

    def formatUUID(self, uuid):
        return dump(uuid)

    # Actual actions
    def getPartitionRowList(self, params):
        """
          Get a list of partition rows, bounded by min & max and involving
          given node.
          Parameters: [min [max [node]]]
            min: offset of the first row to fetch (starts at 0)
            max: offset of the last row to fetch (0 for no limit)
            node: filters the list of nodes serving a line to this node
        """
        params = params + [0, 0, None][len(params):]
        min_offset, max_offset, node = params
        min_offset = int(min_offset)
        max_offset = int(max_offset)
        if node is not None:
            node = self.asNode(node)
        ptid, row_list = self.neoctl.getPartitionRowList(
                min_offset=min_offset, max_offset=max_offset, node=node)
        # TODO: return ptid
        return self.formatRowList(row_list)

    def getNodeList(self, params):
        """
          Get a list of nodes, filtering with given type.
          Parameters: [type]
            type: type of node to display
        """
        assert len(params) < 2
        if len(params):
            node_type = self.asNodeType(params[0])
        else:
            node_type = None
        node_list = self.neoctl.getNodeList(node_type=node_type)
        return self.formatNodeList(node_list)

    def getClusterState(self, params):
        """
          Get cluster state.
        """
        assert len(params) == 0
        return str(self.neoctl.getClusterState())

    def setNodeState(self, params):
        """
          Set node state, and allow (or not) updating partition table.
          Parameters: node state [update]
            node: node to modify
            state: state to put the node in
            update: disallow (0, default) or allow (other integer) partition
                    table to be updated
        """
        assert len(params) in (2, 3)
        node = self.asNode(params[0])
        state = self.asNodeState(params[1])
        if len(params) == 3:
            update_partition_table = bool(int(params[2]))
        else:
            update_partition_table = False
        return self.neoctl.setNodeState(node, state,
            update_partition_table=update_partition_table)

    def setClusterState(self, params):
        """
          Set cluster state.
          Parameters: state
            state: state to put the cluster in
        """
        assert len(params) == 1
        return self.neoctl.setClusterState(self.asClusterState(params[0]))

    def startCluster(self, params):
        """
          Starts cluster operation after a startup.
          Equivalent to:
            set cluster verifying
        """
        assert len(params) == 0
        return self.neoctl.startCluster()

    def enableStorageList(self, params):
        """
          Enable cluster to make use of pending storages.
          Parameters: all
                      node [node [...]]
            node: if "all", add all pending storage nodes.
                  otherwise, the list of storage nodes to enable.
        """
        if len(params) == 1 and params[0] == 'all':
            node_list = self.neoctl.getNodeList(NodeTypes.STORAGE)
            uuid_list = [node[2] for node in node_list]
        else:
            uuid_list = [self.asNode(x) for x in params]
        return self.neoctl.enableStorageList(uuid_list)

    def dropNode(self, params):
        """
          Set node into DOWN state.
          Parameters: node
            node: node the pu into DOWN state
          Equivalent to:
            set node state (node) DOWN
        """
        assert len(params) == 1
        return self.neoctl.dropNode(self.asNode(params[0]))

    def getPrimary(self, params):
        """
          Get primary master node.
        """
        return self.formatUUID(self.neoctl.getPrimary())

class Application(object):
    """The storage node application."""

    def __init__(self, address):
        self.neoctl = TerminalNeoCTL(address)

    def execute(self, args):
        """Execute the command given."""
        # print node type : print list of node of the given type
        # (STORAGE_NODE_TYPE, MASTER_NODE_TYPE...)
        # set node uuid state [1|0] : set the node for the given uuid to the
        # state (RUNNING, DOWN...) and modify the partition if asked
        # set cluster name [shutdown|operational] : either shutdown the
        # cluster or mark it as operational
        current_action = action_dict
        level = 0
        while current_action is not None and \
              level < len(args) and \
              isinstance(current_action, dict):
            current_action = current_action.get(args[level])
            level += 1
        action = None
        if isinstance(current_action, basestring):
            action = getattr(self.neoctl, current_action, None)
        if action is None:
            return self.usage('unknown command')
        try:
            return action(args[level:])
        except NotReadyException, message:
            return 'ERROR: %s' % (message, )

    def _usage(self, action_dict, level=0):
        result = []
        append = result.append
        sub_level = level + 1
        for name, action in action_dict.iteritems():
            append('%s%s' % ('  ' * level, name))
            if isinstance(action, dict):
                append(self._usage(action, level=sub_level))
            else:
                real_action = getattr(self.neoctl, action, None)
                if real_action is None:
                    continue
                docstring = getattr(real_action, '__doc__', None)
                if docstring is None:
                    docstring = '(no docstring)'
                docstring_line_list = docstring.split('\n')
                # Strip empty lines at begining & end of line list
                for end in (0, -1):
                    while len(docstring_line_list) \
                          and docstring_line_list[end] == '':
                        docstring_line_list.pop(end)
                # Get the indentation of first line, to preserve other lines
                # relative indentation.
                first_line = docstring_line_list[0]
                base_indentation = len(first_line) - len(first_line.lstrip())
                result.extend([('  ' * sub_level) + x[base_indentation:] \
                               for x in docstring_line_list])
        return '\n'.join(result)

    def usage(self, message):
        output_list = [message, 'Available commands:', self._usage(action_dict)]
        return '\n'.join(output_list)




#
# (c) Dave Kirby 2001 - 2005
#     mock@thedeveloperscoach.com
#
# Original call interceptor and call assertion code by Phil Dawes (pdawes@users.sourceforge.net)
# Call interceptor code enhanced by Bruce Cropley (cropleyb@yahoo.com.au)
#
# This Python  module and associated files are released under the FreeBSD
# license. Essentially, you can do what you like with it except pretend you wrote
# it yourself.
# 
# 
#     Copyright (c) 2005, Dave Kirby
# 
#     All rights reserved.
# 
#     Redistribution and use in source and binary forms, with or without
#     modification, are permitted provided that the following conditions are met:
# 
#         * Redistributions of source code must retain the above copyright
#           notice, this list of conditions and the following disclaimer.
# 
#         * Redistributions in binary form must reproduce the above copyright
#           notice, this list of conditions and the following disclaimer in the
#           documentation and/or other materials provided with the distribution.
# 
#         * Neither the name of this library nor the names of its
#           contributors may be used to endorse or promote products derived from
#           this software without specific prior written permission.
# 
#     THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
#     ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
#     WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#     DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
#     ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
#     (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
#     LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
#     ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
#     (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
#     SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# 
#         mock@thedeveloperscoach.com


"""
Mock object library for Python. Mock objects can be used when unit testing
to remove a dependency on another production class. They are typically used
when the dependency would either pull in lots of other classes, or
significantly slow down the execution of the test.
They are also used to create exceptional conditions that cannot otherwise
be easily triggered in the class under test.
"""

__version__ = "0.1.0"

# Added in Python 2.1
import inspect
import re

class MockInterfaceError(Exception):
    pass

class Mock:
    """
    The Mock class emulates any other class for testing purposes.
    All method calls are stored for later examination.
    """

    def __init__(self, returnValues=None, realClass=None):
        """
        The Mock class constructor takes a dictionary of method names and
        the values they return.  Methods that are not in the returnValues
        dictionary will return None.
        You may also supply a class whose interface is being mocked.
        All calls will be checked to see if they appear in the original
        interface. Any calls to methods not appearing in the real class
        will raise a MockInterfaceError.  Any calls that would fail due to
        non-matching parameter lists will also raise a MockInterfaceError.
        Both of these help to prevent the Mock class getting out of sync
        with the class it is Mocking.
        """
        self.mockCalledMethods = {}
        self.mockAllCalledMethods = []
        self.mockReturnValues = returnValues or {}
        self.mockExpectations = {}
        self.realClassMethods = None
        if realClass:
            self.realClassMethods = dict(inspect.getmembers(realClass, inspect.isroutine))
            for retMethod in self.mockReturnValues.keys():
                if not self.realClassMethods.has_key(retMethod):
                    raise MockInterfaceError("Return value supplied for method '%s' that was not in the original class" % retMethod)
        self._setupSubclassMethodInterceptors()
     
    def _setupSubclassMethodInterceptors(self):
        methods = inspect.getmembers(self.__class__,inspect.isroutine)
        baseMethods = dict(inspect.getmembers(Mock, inspect.ismethod))
        for m in methods:
            name = m[0]
            # Don't record calls to methods of Mock base class.
            if not name in baseMethods:
                self.__dict__[name] = MockCallable(name, self, handcrafted=True)
 
    def __getattr__(self, name):
        return MockCallable(name, self)
    
    def mockAddReturnValues(self, **methodReturnValues ):
        self.mockReturnValues.update(methodReturnValues)
        
    def mockSetExpectation(self, name, testFn, after=0, until=0):
        self.mockExpectations.setdefault(name, []).append((testFn,after,until))

    def _checkInterfaceCall(self, name, callParams, callKwParams):
        """
        Check that a call to a method of the given name to the original
        class with the given parameters would not fail. If it would fail,
        raise a MockInterfaceError.
        Based on the Python 2.3.3 Reference Manual section 5.3.4: Calls.
        """
        if self.realClassMethods == None:
            return
        if not self.realClassMethods.has_key(name):
            raise MockInterfaceError("Calling mock method '%s' that was not found in the original class" % name)

        func = self.realClassMethods[name]
        try:
            args, varargs, varkw, defaults = inspect.getargspec(func)
        except TypeError:
            # func is not a Python function. It is probably a builtin,
            # such as __repr__ or __coerce__. TODO: Checking?
            # For now assume params are OK.
            return

        # callParams doesn't include self; args does include self.
        numPosCallParams = 1 + len(callParams)

        if numPosCallParams > len(args) and not varargs:
            raise MockInterfaceError("Original %s() takes at most %s arguments (%s given)" % 
                (name, len(args), numPosCallParams))

        # Get the number of positional arguments that appear in the call,
        # also check for duplicate parameters and unknown parameters
        numPosSeen = _getNumPosSeenAndCheck(numPosCallParams, callKwParams, args, varkw)

        lenArgsNoDefaults = len(args) - len(defaults or [])
        if numPosSeen < lenArgsNoDefaults:
            raise MockInterfaceError("Original %s() takes at least %s arguments (%s given)" % (name, lenArgsNoDefaults, numPosSeen))

    def mockGetAllCalls(self):
        """
        Return a list of MockCall objects,
        representing all the methods in the order they were called.
        """
        return self.mockAllCalledMethods
    getAllCalls = mockGetAllCalls  # deprecated - kept for backward compatibility

    def mockGetNamedCalls(self, methodName):
        """
        Return a list of MockCall objects,
        representing all the calls to the named method in the order they were called.
        """
        return self.mockCalledMethods.get(methodName, [])
    getNamedCalls = mockGetNamedCalls  # deprecated - kept for backward compatibility

    def mockCheckCall(self, index, name, *args, **kwargs):
        '''test that the index-th call had the specified name and parameters'''
        call = self.mockAllCalledMethods[index]
        assert name == call.getName(), "%r != %r" % (name, call.getName())
        call.checkArgs(*args, **kwargs)


def _getNumPosSeenAndCheck(numPosCallParams, callKwParams, args, varkw):
    """
    Positional arguments can appear as call parameters either named as
    a named (keyword) parameter, or just as a value to be matched by
    position. Count the positional arguments that are given by either
    keyword or position, and check for duplicate specifications.
    Also check for arguments specified by keyword that do not appear
    in the method's parameter list.
    """
    posSeen = {}
    for arg in args[:numPosCallParams]:
        posSeen[arg] = True
    for kwp in callKwParams:
        if posSeen.has_key(kwp):
            raise MockInterfaceError("%s appears as both a positional and named parameter." % kwp)
        if kwp in args:
            posSeen[kwp] = True
        elif not varkw:
            raise MockInterfaceError("Original method does not have a parameter '%s'" % kwp)
    return len(posSeen)

class MockCall:
    """
    MockCall records the name and parameters of a call to an instance
    of a Mock class. Instances of MockCall are created by the Mock class,
    but can be inspected later as part of the test.
    """
    def __init__(self, name, params, kwparams ):
        self.name = name
        self.params = params
        self.kwparams = kwparams

    def checkArgs(self, *args, **kwargs):
        assert args == self.params, "%r != %r" % (args, self.params)
        assert kwargs == self.kwparams, "%r != %r" % (kwargs, self.kwparams)

    def getParam( self, n ):
        if isinstance(n, int):
            return self.params[n]
        elif isinstance(n, str):
            return self.kwparams[n]
        else:
            raise IndexError, 'illegal index type for getParam'

    def getNumParams(self):
        return len(self.params)

    def getNumKwParams(self):
        return len(self.kwparams)

    def getName(self):
        return self.name
    
    #pretty-print the method call
    def __str__(self):
        s = self.name + "("
        sep = ''
        for p in self.params:
            s = s + sep + repr(p)
            sep = ', '
        items = self.kwparams.items()
        items.sort()
        for k,v in items:
            s = s + sep + k + '=' + repr(v)
            sep = ', '
        s = s + ')'
        return s
    def __repr__(self):
        return self.__str__()

class MockCallable:
    """
    Intercepts the call and records it, then delegates to either the mock's
    dictionary of mock return values that was passed in to the constructor,
    or a handcrafted method of a Mock subclass.
    """
    def __init__(self, name, mock, handcrafted=False):
        self.name = name
        self.mock = mock
        self.handcrafted = handcrafted

    def __call__(self,  *params, **kwparams):
        self.mock._checkInterfaceCall(self.name, params, kwparams)
        thisCall = self.recordCall(params,kwparams)
        self.checkExpectations(thisCall, params, kwparams)
        return self.makeCall(params, kwparams)

    def recordCall(self, params, kwparams):
        """
        Record the MockCall in an ordered list of all calls, and an ordered
        list of calls for that method name.
        """
        thisCall = MockCall(self.name, params, kwparams)
        calls = self.mock.mockCalledMethods.setdefault(self.name, [])
        calls.append(thisCall)
        self.mock.mockAllCalledMethods.append(thisCall)
        return thisCall

    def makeCall(self, params, kwparams):
        if self.handcrafted:
            allPosParams = (self.mock,) + params
            func = _findFunc(self.mock.__class__, self.name)
            if not func:
                raise NotImplementedError
            return func(*allPosParams, **kwparams)
        else:
            returnVal = self.mock.mockReturnValues.get(self.name)
            if isinstance(returnVal, ReturnValuesBase):
                returnVal = returnVal.next()
            return returnVal

    def checkExpectations(self, thisCall, params, kwparams):
        if self.name in self.mock.mockExpectations:
            callsMade = len(self.mock.mockCalledMethods[self.name])
            for (expectation, after, until) in self.mock.mockExpectations[self.name]:
                if callsMade > after and (until==0 or callsMade < until):
                    assert expectation(self.mock, thisCall, len(self.mock.mockAllCalledMethods)-1), 'Expectation failed: '+str(thisCall)


def _findFunc(cl, name):
    """ Depth first search for a method with a given name. """
    if cl.__dict__.has_key(name):
        return cl.__dict__[name]
    for base in cl.__bases__:
        func = _findFunc(base, name)
        if func:
            return func
    return None



class ReturnValuesBase:
    def next(self):
        try:
            return self.iter.next()
        except StopIteration:
            raise AssertionError("No more return values")
    def __iter__(self):
        return self

class ReturnValues(ReturnValuesBase):
    def __init__(self, *values):
        self.iter = iter(values)
        

class ReturnIterator(ReturnValuesBase):
    def __init__(self, iterator):
        self.iter = iter(iterator)

        
def expectParams(*params, **keywords):
    '''check that the callObj is called with specified params and keywords
    '''
    def fn(mockObj, callObj, idx):
        return callObj.params == params and callObj.kwparams == keywords
    return fn


def expectAfter(*methods):
    '''check that the function is only called after all the functions in 'methods'
    '''
    def fn(mockObj, callObj, idx):
        calledMethods = [method.getName() for method in mockObj.mockGetAllCalls()]
        #skip last entry, since that is the current call
        calledMethods = calledMethods[:-1]
        for method in methods:  
            if method not in calledMethods:
                return False
        return True
    return fn

def expectException(exception, *args, **kwargs):
    ''' raise an exception when the method is called
    '''
    def fn(mockObj, callObj, idx):
        raise exception(*args, **kwargs)
    return fn


def expectParam(paramIdx, cond):
    '''check that the callObj is called with parameter specified by paramIdx (a position index or keyword)
    fulfills the condition specified by cond.
    cond is a function that takes a single argument, the value to test.
    '''
    def fn(mockObj, callObj, idx):
        param = callObj.getParam(paramIdx)
        return cond(param)
    return fn

def EQ(value):
    def testFn(param):
        return param == value
    return testFn

def NE(value):
    def testFn(param):
        return param != value
    return testFn

def GT(value):
    def testFn(param):
        return param > value
    return testFn

def LT(value):
    def testFn(param):
        return param < value
    return testFn

def GE(value):
    def testFn(param):
        return param >= value
    return testFn

def LE(value):
    def testFn(param):
        return param <= value
    return testFn

def AND(*condlist):
    def testFn(param):
        for cond in condlist:
            if not cond(param):
                return False
        return True
    return testFn

def OR(*condlist):
    def testFn(param):
        for cond in condlist:
            if cond(param):
                return True
        return False
    return testFn

def NOT(cond):
    def testFn(param):
        return not cond(param)
    return testFn

def MATCHES(regex, *args, **kwargs):
    compiled_regex = re.compile(regex, *args, **kwargs)
    def testFn(param):
        return compiled_regex.match(param) != None
    return testFn

def SEQ(*sequence):
    iterator = iter(sequence)
    def testFn(param):
        try:
            cond = iterator.next()
        except StopIteration:
            raise AssertionError('SEQ exhausted')
        return cond(param)
    return testFn

def IS(instance):
    def testFn(param):
        return param is instance
    return testFn

def ISINSTANCE(class_):
    def testFn(param):
        return isinstance(param, class_) 
    return testFn

def ISSUBCLASS(class_):
    def testFn(param):
        return issubclass(param, class_) 
    return testFn

def CONTAINS(val):
    def testFn(param):
        return val in param 
    return testFn

def IN(container):
    def testFn(param):
        return param in container
    return testFn

def HASATTR(attr):
    def testFn(param):
        return hasattr(param, attr)
    return testFn

def HASMETHOD(method):
    def testFn(param):
        return hasattr(param, method) and callable(getattr(param, method))
    return testFn

CALLABLE = callable






#
# Copyright (C) 2006-2010  Nexedi SA

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.

import neo.lib

from neo.lib.protocol import ProtocolError
from neo.lib.protocol import Packets
from neo.master.handlers import BaseServiceHandler
from neo.lib.exception import OperationFailure
from neo.lib.util import dump
from neo.lib.connector import ConnectorConnectionClosedException
from neo.lib.pt import PartitionTableException


class StorageServiceHandler(BaseServiceHandler):
    """ Handler dedicated to storages during service state """

    def connectionCompleted(self, conn):
        # TODO: unit test
        app = self.app
        uuid = conn.getUUID()
        node = app.nm.getByUUID(uuid)
        app.setStorageNotReady(uuid)
        # XXX: what other values could happen ?
        if node.isRunning():
            conn.notify(Packets.StartOperation())

    def nodeLost(self, conn, node):
        neo.lib.logging.info('storage node lost')
        assert not node.isRunning(), node.getState()

        if not self.app.pt.operational():
            raise OperationFailure, 'cannot continue operation'
        # this is intentionaly placed after the raise because the last cell in a
        # partition must not oudated to allows a cluster restart.
        self.app.outdateAndBroadcastPartition()
        self.app.tm.forget(conn.getUUID())
        if self.app.packing is not None:
            self.answerPack(conn, False)

    def askLastIDs(self, conn):
        app = self.app
        loid = app.tm.getLastOID()
        ltid = app.tm.getLastTID()
        conn.answer(Packets.AnswerLastIDs(loid, ltid, app.pt.getID()))

    def askUnfinishedTransactions(self, conn):
        tm = self.app.tm
        pending_list = tm.registerForNotification(conn.getUUID())
        last_tid = tm.getLastTID()
        p = Packets.AnswerUnfinishedTransactions(last_tid, pending_list)
        conn.answer(p)

    def answerInformationLocked(self, conn, ttid):
        tm = self.app.tm
        if ttid not in tm:
            raise ProtocolError('Unknown transaction')
        # transaction locked on this storage node
        self.app.tm.lock(ttid, conn.getUUID())

    def notifyReplicationDone(self, conn, offset):
        node = self.app.nm.getByUUID(conn.getUUID())
        neo.lib.logging.debug("%s is up for offset %s" % (node, offset))
        try:
            cell_list = self.app.pt.setUpToDate(node, offset)
        except PartitionTableException, e:
            raise ProtocolError(str(e))
        self.app.broadcastPartitionChanges(cell_list)

    def answerPack(self, conn, status):
        app = self.app
        if app.packing is not None:
            client, msg_id, uid_set = app.packing
            uid_set.remove(conn.getUUID())
            if not uid_set:
                app.packing = None
                if not client.isClosed():
                    client.answer(Packets.AnswerPack(True), msg_id=msg_id)




#
# Copyright (C) 2009-2010  Nexedi SA
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.

import unittest
import socket
from neo.lib.protocol import NodeTypes, NodeStates, CellStates, ClusterStates
from neo.lib.protocol import ErrorCodes, Packets, Errors, LockState
from neo.tests import NeoUnitTestBase, IP_VERSION_FORMAT_DICT

class ProtocolTests(NeoUnitTestBase):

    def setUp(self):
        NeoUnitTestBase.setUp(self)
        self.ltid = None

    def getNextTID(self):
        self.ltid = super(ProtocolTests, self).getNextTID(self.ltid)
        return self.ltid

    def test_03_protocolError(self):
        p = Errors.ProtocolError("bad protocol")
        error_code, error_msg = p.decode()
        self.assertEqual(error_code, ErrorCodes.PROTOCOL_ERROR)
        self.assertEqual(error_msg, "bad protocol")

    def test_05_notReady(self):
        p = Errors.NotReady("wait")
        error_code, error_msg = p.decode()
        self.assertEqual(error_code, ErrorCodes.NOT_READY)
        self.assertEqual(error_msg, "wait")

    def test_06_brokenNodeDisallowedError(self):
        p = Errors.BrokenNode("broken")
        error_code, error_msg = p.decode()
        self.assertEqual(error_code, ErrorCodes.BROKEN_NODE)
        self.assertEqual(error_msg, "broken")

    def test_07_oidNotFound(self):
        p = Errors.OidNotFound("no oid")
        error_code, error_msg = p.decode()
        self.assertEqual(error_code, ErrorCodes.OID_NOT_FOUND)
        self.assertEqual(error_msg, "no oid")

    def test_08_tidNotFound(self):
        p = Errors.TidNotFound("no tid")
        error_code, error_msg = p.decode()
        self.assertEqual(error_code, ErrorCodes.TID_NOT_FOUND)
        self.assertEqual(error_msg, "no tid")

    def test_09_ping(self):
        p = Packets.Ping()
        self.assertEqual(p.decode(), ())

    def test_10_pong(self):
        p = Packets.Pong()
        self.assertEqual(p.decode(), ())

    def test_11_RequestIdentification(self):
        uuid = self.getNewUUID()
        p = Packets.RequestIdentification(NodeTypes.CLIENT,
                uuid, (self.local_ip, 9080), "unittest")
        node, p_uuid, (ip, port), name  = p.decode()
        self.assertEqual(node, NodeTypes.CLIENT)
        self.assertEqual(p_uuid, uuid)
        self.assertEqual(ip, self.local_ip)
        self.assertEqual(port, 9080)
        self.assertEqual(name, "unittest")

    def test_11_bis_RequestIdentification_IPv6(self):
        uuid = self.getNewUUID()
        self.local_ip = IP_VERSION_FORMAT_DICT[socket.AF_INET6]
        p = Packets.RequestIdentification(NodeTypes.CLIENT,
                uuid, (self.local_ip, 9080), "unittest")
        node, p_uuid, (ip, port), name  = p.decode()
        self.assertEqual(node, NodeTypes.CLIENT)
        self.assertEqual(p_uuid, uuid)
        self.assertEqual(ip, self.local_ip)
        self.assertEqual(port, 9080)
        self.assertEqual(name, "unittest")

    def test_12_AcceptIdentification(self):
        uuid1, uuid2 = self.getNewUUID(), self.getNewUUID()
        p = Packets.AcceptIdentification(NodeTypes.CLIENT, uuid1,
            10, 20, uuid2)
        node, p_uuid, nb_partitions, nb_replicas, your_uuid  = p.decode()
        self.assertEqual(node, NodeTypes.CLIENT)
        self.assertEqual(p_uuid, uuid1)
        self.assertEqual(nb_partitions, 10)
        self.assertEqual(nb_replicas, 20)
        self.assertEqual(your_uuid, uuid2)

    def test_13_askPrimary(self):
        p = Packets.AskPrimary()
        self.assertEqual(p.decode(), ())

    def test_14_answerPrimary(self):
        uuid = self.getNewUUID()
        uuid1 = self.getNewUUID()
        uuid2 = self.getNewUUID()
        uuid3 = self.getNewUUID()
        master_list = [(("127.0.0.1", 1), uuid1),
                       (("127.0.0.2", 2), uuid2),
                       (("127.0.0.3", 3), uuid3)]
        p = Packets.AnswerPrimary(uuid, master_list)
        primary_uuid, p_master_list  = p.decode()
        self.assertEqual(primary_uuid, uuid)
        self.assertEqual(master_list, p_master_list)

    def test_14_bis_answerPrimaryIPv6(self):
        """ Try to get primary master through IPv6 """
        self.address_type = socket.AF_INET6
        uuid = self.getNewUUID()
        uuid1 = self.getNewUUID()
        uuid2 = self.getNewUUID()
        uuid3 = self.getNewUUID()
        master_list = [(("::1", 1), uuid1),
                       (("::2", 2), uuid2),
                       (("::3", 3), uuid3)]
        p = Packets.AnswerPrimary(uuid, master_list)
        primary_uuid, p_master_list  = p.decode()
        self.assertEqual(primary_uuid, uuid)
        self.assertEqual(master_list, p_master_list)

    def test_15_announcePrimary(self):
        p = Packets.AnnouncePrimary()
        self.assertEqual(p.decode(), ())

    def test_16_reelectPrimary(self):
        p = Packets.ReelectPrimary()
        self.assertEqual(p.decode(), ())

    def test_17_notifyNodeInformation(self):
        uuid1 = self.getNewUUID()
        uuid2 = self.getNewUUID()
        uuid3 = self.getNewUUID()
        node_list = \
            [(NodeTypes.CLIENT, ("127.0.0.1", 1), uuid1, NodeStates.RUNNING),
             (NodeTypes.CLIENT, ("127.0.0.2", 2), uuid2, NodeStates.DOWN),
             (NodeTypes.CLIENT, ("127.0.0.3", 3), uuid3, NodeStates.BROKEN
            )]
        p = Packets.NotifyNodeInformation(node_list)
        p_node_list = p.decode()[0]
        self.assertEqual(node_list, p_node_list)

    def test_18_askLastIDs(self):
        p = Packets.AskLastIDs()
        self.assertEqual(p.decode(), ())

    def test_19_answerLastIDs(self):
        oid = self.getNextTID()
        tid = self.getNextTID()
        ptid = self.getPTID()
        p = Packets.AnswerLastIDs(oid, tid, ptid)
        loid, ltid, lptid = p.decode()
        self.assertEqual(loid, oid)
        self.assertEqual(ltid, tid)
        self.assertEqual(lptid, ptid)

    def test_20_askPartitionTable(self):
        self.assertEqual(Packets.AskPartitionTable().decode(), ())

    def test_21_answerPartitionTable(self):
        ptid = self.getPTID()
        uuid1 = self.getNewUUID()
        uuid2 = self.getNewUUID()
        uuid3 = self.getNewUUID()
        cell_list = [
            (0, [(uuid1, CellStates.UP_TO_DATE), (uuid2, CellStates.OUT_OF_DATE)]),
            (43, [(uuid2, CellStates.OUT_OF_DATE), (uuid3, CellStates.DISCARDED)]),
            (124, [(uuid1, CellStates.DISCARDED), (uuid3, CellStates.UP_TO_DATE)]),
        ]
        p = Packets.AnswerPartitionTable(ptid, cell_list)
        pptid, p_cell_list  = p.decode()
        self.assertEqual(pptid, ptid)
        self.assertEqual(p_cell_list, cell_list)

    def test_22_sendPartitionTable(self):
        ptid = self.getPTID()
        uuid1 = self.getNewUUID()
        uuid2 = self.getNewUUID()
        uuid3 = self.getNewUUID()
        cell_list = [
            (0, [(uuid1, CellStates.UP_TO_DATE), (uuid2, CellStates.OUT_OF_DATE)]),
            (43, [(uuid2, CellStates.OUT_OF_DATE), (uuid3, CellStates.DISCARDED)]),
            (124, [(uuid1, CellStates.DISCARDED), (uuid3, CellStates.UP_TO_DATE)]),
        ]
        p = Packets.AnswerPartitionTable(ptid, cell_list)
        pptid, p_cell_list  = p.decode()
        self.assertEqual(pptid, ptid)
        self.assertEqual(p_cell_list, cell_list)

    def test_23_notifyPartitionChanges(self):
        ptid = self.getPTID()
        uuid1 = self.getNewUUID()
        uuid2 = self.getNewUUID()
        cell_list = [(0, uuid1, CellStates.UP_TO_DATE),
                     (43, uuid2, CellStates.OUT_OF_DATE),
                     (124, uuid1, CellStates.DISCARDED)]
        p = Packets.NotifyPartitionChanges(ptid, cell_list)
        pptid, p_cell_list  = p.decode()
        self.assertEqual(pptid, ptid)
        self.assertEqual(p_cell_list, cell_list)

    def test_24_startOperation(self):
        p = Packets.StartOperation()
        self.assertEqual(p.decode(), ())

    def test_25_stopOperation(self):
        p = Packets.StopOperation()
        self.assertEqual(p.decode(), ())

    def test_26_askUnfinishedTransaction(self):
        p = Packets.AskUnfinishedTransactions()
        self.assertEqual(p.decode(), ())

    def test_27_answerUnfinishedTransaction(self):
        tid = self.getNextTID()
        tid1 = self.getNextTID()
        tid2 = self.getNextTID()
        tid3 = self.getNextTID()
        tid4 = self.getNextTID()
        tid_list = [tid1, tid2, tid3, tid4]
        p = Packets.AnswerUnfinishedTransactions(tid, tid_list)
        p_tid, p_tid_list  = p.decode()
        self.assertEqual(p_tid, tid)
        self.assertEqual(p_tid_list, tid_list)

    def test_28_askObjectPresent(self):
        oid = self.getNextTID()
        tid = self.getNextTID()
        p = Packets.AskObjectPresent(oid, tid)
        loid, ltid = p.decode()
        self.assertEqual(loid, oid)
        self.assertEqual(ltid, tid)

    def test_29_answerObjectPresent(self):
        oid = self.getNextTID()
        tid = self.getNextTID()
        p = Packets.AnswerObjectPresent(oid, tid)
        loid, ltid = p.decode()
        self.assertEqual(loid, oid)
        self.assertEqual(ltid, tid)

    def test_30_deleteTransaction(self):
        tid = self.getNextTID()
        oid_list = [self.getOID(1), self.getOID(2)]
        p = Packets.DeleteTransaction(tid, oid_list)
        self.assertEqual(type(p), Packets.DeleteTransaction)
        self.assertEqual(p.decode(), (tid, oid_list))

    def test_31_commitTransaction(self):
        tid = self.getNextTID()
        p = Packets.CommitTransaction(tid)
        ptid = p.decode()[0]
        self.assertEqual(ptid, tid)

    def test_32_askBeginTransaction(self):
        tid = self.getNextTID()
        p = Packets.AskBeginTransaction(tid)
        ptid = p.decode()[0]
        self.assertEqual(tid, ptid)

    def test_33_answerBeginTransaction(self):
        tid = self.getNextTID()
        p = Packets.AnswerBeginTransaction(tid)
        ptid = p.decode()[0]
        self.assertEqual(ptid, tid)

    def test_34_askNewOIDs(self):
        p = Packets.AskNewOIDs(10)
        nb = p.decode()
        self.assertEqual(nb, (10,))

    def test_35_answerNewOIDs(self):
        oid1 = self.getNextTID()
        oid2 = self.getNextTID()
        oid3 = self.getNextTID()
        oid4 = self.getNextTID()
        oid_list = [oid1, oid2, oid3, oid4]
        p = Packets.AnswerNewOIDs(oid_list)
        p_oid_list  = p.decode()[0]
        self.assertEqual(p_oid_list, oid_list)

    def test_36_askFinishTransaction(self):
        self._testXIDAndYIDList(Packets.AskFinishTransaction)

    def _testXIDAndYIDList(self, packet):
        oid1 = self.getNextTID()
        oid2 = self.getNextTID()
        oid3 = self.getNextTID()
        oid4 = self.getNextTID()
        tid = self.getNextTID()
        oid_list = [oid1, oid2, oid3, oid4]
        p = packet(tid, oid_list)
        p_tid, p_oid_list  = p.decode()
        self.assertEqual(p_tid, tid)
        self.assertEqual(p_oid_list, oid_list)

    def test_37_answerTransactionFinished(self):
        ttid = self.getNextTID()
        tid = self.getNextTID()
        p = Packets.AnswerTransactionFinished(ttid, tid)
        pttid, ptid = p.decode()
        self.assertEqual(pttid, ttid)
        self.assertEqual(ptid, tid)

    def test_38_askLockInformation(self):
        oid1 = self.getNextTID()
        oid2 = self.getNextTID()
        oid_list = [oid1, oid2]
        ttid = self.getNextTID()
        tid = self.getNextTID()
        p = Packets.AskLockInformation(ttid, tid, oid_list)
        pttid, ptid, p_oid_list = p.decode()
        self.assertEqual(pttid, ttid)
        self.assertEqual(ptid, tid)
        self.assertEqual(oid_list, p_oid_list)

    def test_39_answerInformationLocked(self):
        tid = self.getNextTID()
        p = Packets.AnswerInformationLocked(tid)
        ptid = p.decode()[0]
        self.assertEqual(ptid, tid)

    def test_40_invalidateObjects(self):
        oid1 = self.getNextTID()
        oid2 = self.getNextTID()
        oid3 = self.getNextTID()
        oid4 = self.getNextTID()
        tid = self.getNextTID()
        oid_list = [oid1, oid2, oid3, oid4]
        p = Packets.InvalidateObjects(tid, oid_list)
        p_tid, p_oid_list  = p.decode()
        self.assertEqual(p_tid, tid)
        self.assertEqual(p_oid_list, oid_list)

    def test_41_notifyUnlockInformation(self):
        tid = self.getNextTID()
        p = Packets.NotifyUnlockInformation(tid)
        ptid = p.decode()[0]
        self.assertEqual(ptid, tid)

    def test_42_abortTransaction(self):
        tid = self.getNextTID()
        p = Packets.AbortTransaction(tid)
        ptid = p.decode()[0]
        self.assertEqual(ptid, tid)

    def test_43_askStoreTransaction(self):
        tid = self.getNextTID()
        oid1 = self.getNextTID()
        oid2 = self.getNextTID()
        oid3 = self.getNextTID()
        oid4 = self.getNextTID()
        oid_list = [oid1, oid2, oid3, oid4]
        p = Packets.AskStoreTransaction(tid, "moi", "transaction", "exti", oid_list)
        ptid, user, desc, ext, p_oid_list = p.decode()
        self.assertEqual(ptid, tid)
        self.assertEqual(p_oid_list, oid_list)
        self.assertEqual(user, "moi")
        self.assertEqual(desc, "transaction")
        self.assertEqual(ext, "exti")

    def test_44_answerStoreTransaction(self):
        tid = self.getNextTID()
        p = Packets.AnswerStoreTransaction(tid)
        ptid = p.decode()[0]
        self.assertEqual(ptid, tid)

    def test_45_askStoreObject(self):
        oid = self.getNextTID()
        serial = self.getNextTID()
        tid = self.getNextTID()
        tid2 = self.getNextTID()
        unlock = False
        p = Packets.AskStoreObject(oid, serial, 1, 55, "to", tid2, tid, unlock)
        poid, pserial, compression, checksum, data, ptid2, ptid, punlock = \
            p.decode()
        self.assertEqual(oid, poid)
        self.assertEqual(serial, pserial)
        self.assertEqual(tid, ptid)
        self.assertEqual(tid2, ptid2)
        self.assertEqual(compression, 1)
        self.assertEqual(checksum, 55)
        self.assertEqual(data, "to")
        self.assertEqual(unlock, punlock)

    def test_46_answerStoreObject(self):
        oid = self.getNextTID()
        serial = self.getNextTID()
        p = Packets.AnswerStoreObject(True, oid, serial)
        conflicting, poid, pserial = p.decode()
        self.assertEqual(oid, poid)
        self.assertEqual(serial, pserial)
        self.assertTrue(conflicting)

    def test_47_askObject(self):
        oid = self.getNextTID()
        serial = self.getNextTID()
        tid = self.getNextTID()
        p = Packets.AskObject(oid, serial, tid)
        poid, pserial, ptid = p.decode()
        self.assertEqual(oid, poid)
        self.assertEqual(serial, pserial)
        self.assertEqual(tid, ptid)

    def test_48_answerObject(self):
        oid = self.getNextTID()
        serial_start = self.getNextTID()
        serial_end = self.getNextTID()
        data_serial = self.getNextTID()
        p = Packets.AnswerObject(oid, serial_start, serial_end, 1, 55, "to",
            data_serial)
        poid, pserial_start, pserial_end, compression, checksum, data, \
            pdata_serial = p.decode()
        self.assertEqual(oid, poid)
        self.assertEqual(serial_start, pserial_start)
        self.assertEqual(serial_end, pserial_end)
        self.assertEqual(compression, 1)
        self.assertEqual(checksum, 55)
        self.assertEqual(data, "to")
        self.assertEqual(pdata_serial, data_serial)

    def test_49_askTIDs(self):
        p = Packets.AskTIDs(1, 10, 5)
        first, last, partition = p.decode()
        self.assertEqual(first, 1)
        self.assertEqual(last, 10)
        self.assertEqual(partition, 5)

    def test_50_answerTIDs(self):
        self._test_AnswerTIDs(Packets.AnswerTIDs)

    def _test_AnswerTIDs(self, packet):
        tid1 = self.getNextTID()
        tid2 = self.getNextTID()
        tid3 = self.getNextTID()
        tid4 = self.getNextTID()
        tid_list = [tid1, tid2, tid3, tid4]
        p = packet(tid_list)
        p_tid_list  = p.decode()[0]
        self.assertEqual(p_tid_list, tid_list)

    def test_51_askTransactionInfomation(self):
        tid = self.getNextTID()
        p = Packets.AskTransactionInformation(tid)
        ptid = p.decode()[0]
        self.assertEqual(tid, ptid)

    def test_52_answerTransactionInformation(self):
        tid = self.getNextTID()
        oid1 = self.getNextTID()
        oid2 = self.getNextTID()
        oid3 = self.getNextTID()
        oid4 = self.getNextTID()
        oid_list = [oid1, oid2, oid3, oid4]
        p = Packets.AnswerTransactionInformation(tid, "moi",
                "transaction", "exti", False, oid_list)
        ptid, user, desc, ext, packed, p_oid_list = p.decode()
        self.assertEqual(ptid, tid)
        self.assertEqual(p_oid_list, oid_list)
        self.assertEqual(user, "moi")
        self.assertEqual(desc, "transaction")
        self.assertEqual(ext, "exti")
        self.assertFalse(packed)

    def test_53_askObjectHistory(self):
        oid = self.getNextTID()
        p = Packets.AskObjectHistory(oid, 1, 10,)
        poid, first, last = p.decode()
        self.assertEqual(first, 1)
        self.assertEqual(last, 10)
        self.assertEqual(poid, oid)

    def test_54_answerObjectHistory(self):
        oid = self.getNextTID()
        hist1 = (self.getNextTID(), 15)
        hist2 = (self.getNextTID(), 353)
        hist3 = (self.getNextTID(), 326)
        hist4 = (self.getNextTID(), 652)
        hist_list = [hist1, hist2, hist3, hist4]
        p = Packets.AnswerObjectHistory(oid, hist_list)
        poid, p_hist_list  = p.decode()
        self.assertEqual(p_hist_list, hist_list)
        self.assertEqual(oid, poid)

    def test_57_notifyReplicationDone(self):
        offset = 10
        p = Packets.NotifyReplicationDone(offset)
        p_offset = p.decode()[0]
        self.assertEqual(p_offset, offset)

    def test_askObjectUndoSerial(self):
        tid = self.getNextTID()
        ltid = self.getNextTID()
        undone_tid = self.getNextTID()
        oid_list = [self.getOID(x) for x in xrange(4)]
        p = Packets.AskObjectUndoSerial(tid, ltid, undone_tid, oid_list)
        ptid, pltid, pundone_tid, poid_list = p.decode()
        self.assertEqual(tid, ptid)
        self.assertEqual(ltid, pltid)
        self.assertEqual(undone_tid, pundone_tid)
        self.assertEqual(oid_list, poid_list)

    def test_answerObjectUndoSerial(self):
        oid1 = self.getNextTID()
        oid2 = self.getNextTID()
        tid1 = self.getNextTID()
        tid2 = self.getNextTID()
        tid3 = self.getNextTID()
        object_tid_dict = {
          oid1: (tid1, tid2, True),
          oid2: (tid3, None, False),
        }
        p = Packets.AnswerObjectUndoSerial(object_tid_dict)
        pobject_tid_dict = p.decode()[0]
        self.assertEqual(object_tid_dict, pobject_tid_dict)

    def test_NotifyLastOID(self):
        oid = self.getOID(1)
        p = Packets.NotifyLastOID(oid)
        self.assertEqual(p.decode(), (oid, ))

    def test_AnswerClusterState(self):
        state = ClusterStates.RUNNING
        p = Packets.AnswerClusterState(state)
        self.assertEqual(p.decode(), (state, ))

    def test_AskClusterState(self):
        p = Packets.AskClusterState()
        self.assertEqual(p.decode(), ())

    def test_NotifyClusterInformation(self):
        state = ClusterStates.RECOVERING
        p = Packets.NotifyClusterInformation(state)
        self.assertEqual(p.decode(), (state, ))

    def test_SetClusterState(self):
        state = ClusterStates.VERIFYING
        p = Packets.SetClusterState(state)
        self.assertEqual(p.decode(), (state, ))

    def test_AnswerNodeInformation(self):
        p = Packets.AnswerNodeInformation()
        self.assertEqual(p.decode(), ())

    def test_AskNodeInformation(self):
        p = Packets.AskNodeInformation()
        self.assertEqual(p.decode(), ())

    def test_AddPendingNodes(self):
        uuid1, uuid2 = self.getNewUUID(), self.getNewUUID()
        p = Packets.AddPendingNodes((uuid1, uuid2))
        self.assertEqual(p.decode(), ([uuid1, uuid2], ))

    def test_SetNodeState(self):
        uuid = self.getNewUUID()
        state = NodeStates.PENDING
        p = Packets.SetNodeState(uuid, state, True)
        self.assertEqual(p.decode(), (uuid, state, True))

    def test_AskNodeList(self):
        node_type = NodeTypes.STORAGE
        p = Packets.AskNodeList(node_type)
        self.assertEqual(p.decode(), (node_type, ))

    def test_AnswerNodeList(self):
        node1 = (NodeTypes.CLIENT, (self.local_ip, 1000),
                self.getNewUUID(), NodeStates.DOWN)
        node2 = (NodeTypes.MASTER, (self.local_ip, 2000),
                self.getNewUUID(), NodeStates.RUNNING)
        p = Packets.AnswerNodeList((node1, node2))
        self.assertEqual(p.decode(), ([node1, node2], ))

    def test_AnswerNodeListIPv6(self):
        self.address_type = socket.AF_INET6
        node1 = (NodeTypes.CLIENT, (self.local_ip, 1000),
                self.getNewUUID(), NodeStates.DOWN)
        node2 = (NodeTypes.MASTER, (self.local_ip, 2000),
                self.getNewUUID(), NodeStates.RUNNING)
        p = Packets.AnswerNodeList((node1, node2))
        self.assertEqual(p.decode(), ([node1, node2], ))

    def test_AskPartitionList(self):
        min_offset = 10
        max_offset = 20
        uuid = self.getNewUUID()
        p = Packets.AskPartitionList(min_offset, max_offset, uuid)
        self.assertEqual(p.decode(), (min_offset, max_offset, uuid))

    def test_AnswerPartitionList(self):
        ptid = self.getPTID(1)
        row_list = [
            (0, [
                (self.getNewUUID(), CellStates.UP_TO_DATE),
                (self.getNewUUID(), CellStates.OUT_OF_DATE),
                ]),
            (1, [
                (self.getNewUUID(), CellStates.FEEDING),
                (self.getNewUUID(), CellStates.DISCARDED),
                ]),
        ]
        p = Packets.AnswerPartitionList(ptid, row_list)
        self.assertEqual(p.decode(), (ptid, row_list))

    def test_AskHasLock(self):
        tid = self.getNextTID()
        oid = self.getNextTID()
        p = Packets.AskHasLock(tid, oid)
        self.assertEqual(p.decode(), (tid, oid))

    def test_AnswerHasLock(self):
        oid = self.getNextTID()
        for lock_state in LockState.itervalues():
            p = Packets.AnswerHasLock(oid, lock_state)
            self.assertEqual(p.decode(), (oid, lock_state))

    def test_Notify(self):
        msg = 'test'
        self.assertEqual(Packets.Notify(msg).decode(), (msg, ))

    def test_AskTIDsFrom(self):
        tid = self.getNextTID()
        tid2 = self.getNextTID()
        p = Packets.AskTIDsFrom(tid, tid2, 1000, [5])
        min_tid, max_tid, length, partition = p.decode()
        self.assertEqual(min_tid, tid)
        self.assertEqual(max_tid, tid2)
        self.assertEqual(length, 1000)
        self.assertEqual(partition, [5])

    def test_AnswerTIDsFrom(self):
        self._test_AnswerTIDs(Packets.AnswerTIDsFrom)

    def test_AskObjectHistoryFrom(self):
        oid = self.getOID(1)
        min_serial = self.getNextTID()
        max_serial = self.getNextTID()
        length = 5
        partition = 4
        p = Packets.AskObjectHistoryFrom(oid, min_serial, max_serial, length,
            partition)
        p_oid, p_min_serial, p_max_serial, p_length, p_partition = p.decode()
        self.assertEqual(p_oid, oid)
        self.assertEqual(p_min_serial, min_serial)
        self.assertEqual(p_max_serial, max_serial)
        self.assertEqual(p_length, length)
        self.assertEqual(p_partition, partition)

    def test_AnswerObjectHistoryFrom(self):
        object_dict = {}
        for int_oid in xrange(4):
            object_dict[self.getOID(int_oid)] = [self.getNextTID() \
                for _ in xrange(5)]
        p = Packets.AnswerObjectHistoryFrom(object_dict)
        p_object_dict = p.decode()[0]
        self.assertEqual(object_dict, p_object_dict)

    def test_AskCheckTIDRange(self):
        min_tid = self.getNextTID()
        max_tid = self.getNextTID()
        length = 2
        partition = 4
        p = Packets.AskCheckTIDRange(min_tid, max_tid, length, partition)
        p_min_tid, p_max_tid, p_length, p_partition = p.decode()
        self.assertEqual(p_min_tid, min_tid)
        self.assertEqual(p_max_tid, max_tid)
        self.assertEqual(p_length, length)
        self.assertEqual(p_partition, partition)

    def test_AnswerCheckTIDRange(self):
        min_tid = self.getNextTID()
        length = 2
        count = 1
        tid_checksum = self.getNewUUID()
        max_tid = self.getNextTID()
        p = Packets.AnswerCheckTIDRange(min_tid, length, count, tid_checksum,
            max_tid)
        p_min_tid, p_length, p_count, p_tid_checksum, p_max_tid = p.decode()
        self.assertEqual(p_min_tid, min_tid)
        self.assertEqual(p_length, length)
        self.assertEqual(p_count, count)
        self.assertEqual(p_tid_checksum, tid_checksum)
        self.assertEqual(p_max_tid, max_tid)

    def test_AskCheckSerialRange(self):
        min_oid = self.getOID(1)
        min_serial = self.getNextTID()
        max_tid = self.getNextTID()
        length = 2
        partition = 4
        p = Packets.AskCheckSerialRange(min_oid, min_serial, max_tid, length,
            partition)
        p_min_oid, p_min_serial, p_max_tid, p_length, p_partition = p.decode()
        self.assertEqual(p_min_oid, min_oid)
        self.assertEqual(p_min_serial, min_serial)
        self.assertEqual(p_max_tid, max_tid)
        self.assertEqual(p_length, length)
        self.assertEqual(p_partition, partition)

    def test_AnswerCheckSerialRange(self):
        min_oid = self.getOID(1)
        min_serial = self.getNextTID()
        length = 2
        count = 1
        oid_checksum = self.getNewUUID()
        max_oid = self.getOID(5)
        tid_checksum = self.getNewUUID()
        max_serial = self.getNextTID()
        p = Packets.AnswerCheckSerialRange(min_oid, min_serial, length, count,
            oid_checksum, max_oid, tid_checksum, max_serial)
        p_min_oid, p_min_serial, p_length, p_count, p_oid_checksum, \
            p_max_oid, p_tid_checksum, p_max_serial = p.decode()
        self.assertEqual(p_min_oid, min_oid)
        self.assertEqual(p_min_serial, min_serial)
        self.assertEqual(p_length, length)
        self.assertEqual(p_count, count)
        self.assertEqual(p_oid_checksum, oid_checksum)
        self.assertEqual(p_max_oid, max_oid)
        self.assertEqual(p_tid_checksum, tid_checksum)
        self.assertEqual(p_max_serial, max_serial)


    def test_AskPack(self):
        tid = self.getNextTID()
        p = Packets.AskPack(tid)
        ptid = p.decode()[0]
        self.assertEqual(ptid, tid)

    def test_AnswerPack(self):
        status = True
        p = Packets.AnswerPack(status)
        pstatus = p.decode()[0]
        self.assertEqual(pstatus, status)

    def test_notifyReady(self):
        p = Packets.NotifyReady()
        self.assertEqual(tuple(), p.decode())

    def test_AskLastTransaction(self):
        Packets.AskLastTransaction()

    def test_AnswerLastTransaction(self):
        tid = self.getNextTID()
        p = Packets.AnswerLastTransaction(tid)
        ptid = p.decode()[0]
        self.assertEqual(ptid, tid)

    def test_AskCheckCurrentSerial(self):
        tid = self.getNextTID()
        serial = self.getNextTID()
        oid = self.getNextTID()
        p = Packets.AskCheckCurrentSerial(tid, serial, oid)
        ptid, pserial, poid = p.decode()
        self.assertEqual(ptid, tid)
        self.assertEqual(pserial, serial)
        self.assertEqual(poid, oid)

if __name__ == '__main__':
    unittest.main()




#
# Copyright (C) 2006-2010  Nexedi SA
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.

import neo.lib

from neo.lib.node import NodeManager
from neo.lib.event import EventManager
from neo.lib.connection import ListeningConnection
from neo.lib.exception import PrimaryFailure
from neo.admin.handler import AdminEventHandler, MasterEventHandler, \
    MasterRequestEventHandler
from neo.lib.connector import getConnectorHandler
from neo.lib.bootstrap import BootstrapManager
from neo.lib.pt import PartitionTable
from neo.lib.protocol import NodeTypes, NodeStates, Packets, Errors
from neo.lib.debug import register as registerLiveDebugger

class Dispatcher:
    """Dispatcher use to redirect master request to handler"""

    def __init__(self):
        # associate conn/message_id to dispatch
        # message to connection
        self.message_table = {}

    def register(self, msg_id, conn, kw=None):
        self.message_table[msg_id] = conn, kw

    def pop(self, msg_id):
        return self.message_table.pop(msg_id)

    def registered(self, msg_id):
        return self.message_table.has_key(msg_id)

    def clear(self):
        """
            Unregister packet expected for a given connection
        """
        self.message_table.clear()

class Application(object):
    """The storage node application."""

    def __init__(self, config):
        # Internal attributes.
        self.em = EventManager()
        self.nm = NodeManager()

        self.name = config.getCluster()
        self.server = config.getBind()

        self.master_addresses, connector_name = config.getMasters()
        self.connector_handler = getConnectorHandler(connector_name)
        neo.lib.logging.debug('IP address is %s, port is %d', *(self.server))

        # The partition table is initialized after getting the number of
        # partitions.
        self.pt = None
        self.uuid = config.getUUID()
        self.primary_master_node = None
        self.request_handler = MasterRequestEventHandler(self)
        self.master_event_handler = MasterEventHandler(self)
        self.dispatcher = Dispatcher()
        self.cluster_state = None
        self.reset()
        registerLiveDebugger(on_log=self.log)

    def close(self):
        self.listening_conn = None
        self.nm.close()
        self.em.close()
        del self.__dict__

    def reset(self):
        self.bootstrapped = False
        self.master_conn = None
        self.master_node = None

    def log(self):
        self.em.log()
        self.nm.log()
        if self.pt is not None:
            self.pt.log()

    def run(self):
        """Make sure that the status is sane and start a loop."""
        if len(self.name) == 0:
            raise RuntimeError, 'cluster name must be non-empty'

        # Make a listening port.
        handler = AdminEventHandler(self)
        self.listening_conn = ListeningConnection(self.em, handler,
            addr=self.server, connector=self.connector_handler())

        while True:
            self.connectToPrimary()
            try:
                while True:
                    self.em.poll(1)
            except PrimaryFailure:
                neo.lib.logging.error('primary master is down')


    def connectToPrimary(self):
        """Find a primary master node, and connect to it.

        If a primary master node is not elected or ready, repeat
        the attempt of a connection periodically.

        Note that I do not accept any connection from non-master nodes
        at this stage."""

        nm = self.nm
        nm.init()
        self.cluster_state = None

        for address in self.master_addresses:
            self.nm.createMaster(address=address)

        # search, find, connect and identify to the primary master
        bootstrap = BootstrapManager(self, self.name, NodeTypes.ADMIN,
                self.uuid, self.server)
        data = bootstrap.getPrimaryConnection(self.connector_handler)
        (node, conn, uuid, num_partitions, num_replicas) = data
        nm.update([(node.getType(), node.getAddress(), node.getUUID(),
                    NodeStates.RUNNING)])
        self.master_node = node
        self.master_conn = conn
        self.uuid = uuid

        if self.pt is None:
            self.pt = PartitionTable(num_partitions, num_replicas)
        elif self.pt.getPartitions() != num_partitions:
            # XXX: shouldn't we recover instead of raising ?
            raise RuntimeError('the number of partitions is inconsistent')
        elif self.pt.getReplicas() != num_replicas:
            # XXX: shouldn't we recover instead of raising ?
            raise RuntimeError('the number of replicas is inconsistent')

        # passive handler
        self.master_conn.setHandler(self.master_event_handler)
        self.master_conn.ask(Packets.AskNodeInformation())
        self.master_conn.ask(Packets.AskPartitionTable())

    def sendPartitionTable(self, conn, min_offset, max_offset, uuid):
        # we have a pt
        self.pt.log()
        row_list = []
        if max_offset == 0:
            max_offset = self.pt.getPartitions()
        try:
            for offset in xrange(min_offset, max_offset):
                row = []
                try:
                    for cell in self.pt.getCellList(offset):
                        if uuid is not None and cell.getUUID() != uuid:
                            continue
                        else:
                            row.append((cell.getUUID(), cell.getState()))
                except TypeError:
                    pass
                row_list.append((offset, row))
        except IndexError:
            p = Errors.ProtocolError('invalid partition table offset')
            conn.notify(p)
            return
        p = Packets.AnswerPartitionList(self.pt.getID(), row_list)
        conn.answer(p)
